"""MultiQC module to parse the output from SNPsplit"""
import logging
import re

import yaml

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        super(MultiqcModule, self).__init__(
            name="SNPsplit",
            anchor="SNPsplit",
            target="SNPsplit",
            href="https://www.bioinformatics.babraham.ac.uk/projects/SNPsplit/",
            info="A tool to determine allele-specific alignments from high-throughput sequencing experiments that have been aligned to N-masked genomes",
            doi="10.12688/f1000research.9037.2",
        )

        self.snpsplit_data = dict()

        # Parse log files generated with newer versions of SNPsplit (YAML)
        for f in self.find_log_files("snpsplit/new", filehandles=True):
            parsed_log = self.parse_new_snpsplit_log(f)
            self._save_parsed(parsed_log, f)

        # Parse log files generated with older versions of SNPsplit
        for f in self.find_log_files("snpsplit/old"):
            parsed_log = self.parse_old_snpsplit_log(f)
            self._save_parsed(parsed_log, f)

        # Filter --ignore-samples
        self.snpsplit_data = self.ignore_samples(self.snpsplit_data)

        if len(self.snpsplit_data) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(self.snpsplit_data)} reports")

        self.write_data_file(self.snpsplit_data, "multiqc_snpsplit")

        self.add_general_stats()
        self.allele_tagging_section()
        self.allele_sorting_section()

    def _save_parsed(self, parsed, f):
        s_name = self.clean_s_name(parsed[0], f)
        if s_name in self.snpsplit_data:
            log.debug(f"Replacing duplicate sample {s_name}")
        self.snpsplit_data[s_name] = parsed[1]
        self.add_data_source(f, s_name=s_name)
        self.add_software_version(parsed[1].get("version"), s_name)

    @staticmethod
    def parse_new_snpsplit_log(f):
        data = next(yaml.load_all(f["f"], Loader=yaml.SafeLoader))
        flat_data = {}
        for k in data:
            for sk in data[k]:
                key = sk
                for prefix in ["PE_", "SE_", "HiC_"]:
                    if sk.startswith(prefix):
                        key = sk[len(prefix) :]
                flat_key = f"{k.lower()}_{key}"
                flat_data[flat_key] = data[k][sk]
        input_fn = data["Meta"]["infile"]
        flat_data["version"] = data["Meta"]["version"]
        return [input_fn, flat_data]

    def parse_old_snpsplit_log(self, f):
        input_fn = None
        d = dict()

        for line in f["f"].splitlines():
            # Parse the sample name
            match = re.match(r"Input file:\W+'([^']+)'", line)
            if match:
                # Report can have two sections with two input files,
                # the second sorting the output from the tagging
                # Just take the first instance, which is the initial input file
                if input_fn is None:
                    input_fn = match.group(1)
                continue

            # Log format 1: XXX description
            regex_patterns = [
                # Allele-tagging report
                ["tagging_g1", r"(\d+) reads were specific for genome 1"],
                ["tagging_g2", r"(\d+) reads were specific for genome 2"],
                ["tagging_unassignable", r"(\d+) reads were unassignable"],
                ["tagging_bizarre", r"(\d+) contained conflicting allele-specific SNPs"],
                ["tagging_unaligned", r"Reads were unaligned and hence skipped: (\d+)"],
                ["tagging_CT_positions_skipped", r"(\d+) reads that were unassignable contained C>T SNPs"],
                # Allele-specific sorting report
                ["sorting_conflicting", r"Reads contained conflicting SNP information:\W+(\d+)"],
            ]
            for k, regex in regex_patterns:
                match = re.match(regex, line)
                if match:
                    d[k] = int(match.group(1))
                    break

            # Format 2: Decription: XXX
            sorting_patterns = [
                # Tagging meta
                ["tagging_SNP_annotation", "SNP annotation file"],
                ["tagging_SNPs_stored", "SNPs stored in total"],
                ["tagging_N_was_known_SNP", "N was present in the list of known SNPs"],
                ["tagging_N_not_known", "N was not present in the list of SNPs"],
                # Allele sorting
                ["sorting_genome1", "Reads were specific for genome 1"],
                ["sorting_genome2", "Reads were specific for genome 2"],
                ["sorting_unassignable", "Reads were unassignable"],
                ["sorting_conflicting", "Read pairs contained conflicting SNP information"],
                # Hi-C data
                ["sorting_genome1_G1_G1", "Read pairs were specific for genome 1 (G1/G1)"],
                ["sorting_genome2_G2_G2", "Read pairs were specific for genome 2 (G2/G2)"],
                ["sorting_unassignable_UA_UA", "Read pairs were unassignable (UA/UA):"],
                ["sorting_G1_UA_total", "Read pairs were a mix of G1 and UA"],
                ["sorting_G2_UA_total", "Read pairs were a mix of G2 and UA"],
                ["sorting_G1_G2_total", "Read pairs were a mix of G1 and G2"],
            ]
            for k, pattern in sorting_patterns:
                if line.startswith(pattern):
                    try:
                        d[k] = int(line.split("\t")[-1].split()[0])
                    except ValueError:
                        d[k] = line.split("\t")[-1].split()[0]
                    break

            if "tagging_N_was_known_SNP" in d and "tagging_N_not_known" in d:
                n_total = d["tagging_N_was_known_SNP"] + d["tagging_N_not_known"]
                d["tagging_percent_N_was_known_SNP"] = (d["tagging_N_was_known_SNP"] / float(n_total)) * 100.0

        return [input_fn, d]

    def add_general_stats(self):
        """Add some columns to the General Statistics table at the top of the report"""
        headers = {
            "tagging_SNP_annotation": {
                "title": "SNP annotation",
                "description": "Annotation file used for differentiating genomes",
                "scale": False,
                "modify": lambda x: f"<code>{x}</code>",
                "hidden": True,
            },
            "tagging_percent_N_was_known_SNP": {
                "title": "% Ns known SNP",
                "description": "Percentage of detected SNPs in the sample that were also present in the annotation",
                "scale": "RdYlGn",
                "format": "{:,.2f}",
                "max": 100,
                "min": 0,
                "suffix": "%",
            },
            "tagging_SNPs_stored": {
                "title": "SNPs stored",
                "description": "Total number of SNPs used for the analysis",
                "scale": "PRGn",
                "format": "{:,.0f}",
                "hidden": True,
            },
        }
        self.general_stats_addcols(self.snpsplit_data, headers)

    def allele_tagging_section(self):
        """Allele-tagging report"""
        cats = {
            "tagging_g1": {"name": "Genome 1"},
            "tagging_g2": {"name": "Genome 2"},
            "tagging_unassignable": {"name": "Not assigned"},
            "tagging_bizarre": {"name": "Conflicting SNPs"},
            "tagging_unaligned": {"name": "Unaligned reads"},
            "tagging_CT_positions_skipped": {"name": "C->T SNP"},
        }

        # Subtract C->T from unassignable
        plot_data = {}
        for s_name, ss_data in self.snpsplit_data.items():
            if "tagging_unassignable" and "tagging_CT_positions_skipped" in ss_data:
                ss_data["tagging_unassignable"] -= ss_data["tagging_CT_positions_skipped"]
            plot_data[s_name] = ss_data

        pconfig = {
            "id": "snpsplit-allele-tagging-plot",
            "title": "SNPsplit: Allele-tagging report",
            "ylab": "Reads",
            "cpswitch_counts_label": "Reads",
        }

        self.add_section(
            name="Allele-tagging report",
            description="Per-sample metrics of how many reads were assigned to each genome.",
            helptext="""
                Allele tagging works on a per-read basis. The results may therefore differ considerably
                from the Allele-specific sorting results if the samples were paired-end or Hi-C samples.

                For single hybrid genomes, Genome 1-specific reads are specific for the reference sequence
                and Genome 2-specific reads are specific for the straing specified with `--strain`.

                For dual hybrid genomes, Genome 1-specific reads are specific for the strain specified
                with `--strain`, and Genome 2-specific reads are specific for the straing specified with `--strain2`.

                Bar graph categories are:

                * `Genome 1`: Reads assigned to Genome 1
                * `Genome 2`: Reads assigned to Genome 2
                * `Not assigned`: Reads don't overlap a SNP
                * `Conflicting SNPs`: Reads contained allele-specific information for both alleles within the same read
                * `Unaligned reads`: Reads aren't aligned
                * `No match`: Reads overlap informative SNPs, but don't contain the expected nucleotide for either genome
                * `C->T SNP`: (Bisulfite sequencing data only) Read SNPs involved some form of C->T transition, rendering it non-informative for allele-assignment
                """,
            plot=bargraph.plot(plot_data, cats, pconfig),
        )

    def allele_sorting_section(self):
        """Allele-specific sorting report"""
        cats = {
            "sorting_genome1": {"name": "Genome 1"},
            "sorting_genome2": {"name": "Genome 2"},
            "sorting_unassignable": {"name": "Not assigned"},
            "sorting_conflicting": {"name": "Conflicting SNPs"},
            "sorting_genome1_G1_G1": {"name": "Genome 1 / Genome 1"},
            "sorting_genome2_G2_G2": {"name": "Genome 2 / Genome 2"},
            "sorting_unassignable_UA_UA": {"name": "Unassignable / Unassignable"},
            "sorting_G1_UA_total": {"name": "Genome 1 / unassignable"},
            "sorting_G2_UA_total": {"name": "Genome 2 / unassignable"},
            "sorting_G1_G2_total": {"name": "Different genomes"},
        }
        # HiC only

        pconfig = {
            "id": "snpsplit-sorting-plot",
            "title": "SNPsplit: Allele-specific sorting",
            "ylab": "Reads",
            "cpswitch_counts_label": "Reads",
        }

        self.add_section(
            name="Allele-specific sorting",
            description="Per-sample metrics of how reads and pairs of reads were sorted into each genome.",
            helptext="""
                Bargraph categories are:

                * `Genome 1`: Reads assigned to Genome 1
                * `Genome 2`: Reads assigned to Genome 2
                * `Not assigned`: Reads don't overlap a SNP
                * `Conflicting SNPs`: Reads contained allele-specific information for both alleles within the same read

                For HiC data, categories are:

                * `Genome 1 / Genome 1`: Pairs with both reads specific to genome 1
                * `Genome 2 / Genome 2`: Pairs with both reads specific to genome 2
                * `Unassignable / Unassignable`: Pairs with both reads unassignable (not overlapping a SNP)
                * `Genome 1 / unassignable`: One paired-end read assigned to Genome 1, one unassignable (doesn't overlap a SNP)
                * `Genome 2 / unassignable`: One paired-end read assigned to Genome 2, one unassignable (doesn't overlap a SNP)
                * `Different genomes`: Paired-end reads assigned to different genomes
                * `Conflicting SNPs`: Reads contained allele-specific information for both alleles within the same read

                Allele-specific sorting takes both reads of read pairs or Hi-C samples into account.

                Note that metrics here may differ from those in the allele-tagging report.
                This occurs when paired-end reads are used, since 'tagging' only one read in
                a pair as arising from one genome can suffice in both reads being sorted there.

                Similarly, if two reads in a pair are tagged as arising from different genomes
                then the pair becomes unassignable.
            """,
            plot=bargraph.plot(self.snpsplit_data, cats, pconfig),
        )



#!/usr/bin/env python
""" MultiQC module to parse output from truvari """
import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Import the truvari submodules
from .bench import BenchSummary

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule, BenchSummary):
    """This is the MultiQC module for truvari."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Truvari",
            anchor="truvari",
            href="https://github.com/ACEnglish/truvari",
            info="is a toolkit for benchmarking, merging, and annotating structural variants",
            doi="https://doi.org/10.1101/2022.02.21.481353",
        )

        # Set up class objects to hold parsed data
        self.general_stats_headers = dict()
        self.general_stats_data = dict()
        n = dict()

        # Call submodule functions
        n["bench"] = self.parse_bench_stats()
        if n["bench"] > 0:
            log.info(f"Found {n['bench']} truvari bench reports")

        # Exit if we didn't find anything
        if sum(n.values()) == 0:
            raise ModuleNoSamplesFound

        # Add to the General Stats table (has to be called once per MultiQC module)
        self.general_stats_addcols(self.general_stats_data, self.general_stats_headers)



""" MultiQC module to parse output from Sambamba """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Import the Sambamba submodules
from .markdup import SambambaMarkdupMixin

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule, SambambaMarkdupMixin):
    """Sambamba has a number of different commands and outputs.
    This MultiQC module supports some but not all. The code for
    each script is split into its own file and adds a section to
    the module output if logs are found."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Sambamba",
            anchor="sambamba",
            href="https://lomereiter.github.io/sambamba/",
            info=" is a suite of programs for interacting with high-throughput sequencing data.",
            doi="10.1093/bioinformatics/btv098",
        )

        # Set up class objects to hold parsed data
        self.general_stats_headers = dict()
        self.general_stats_data = dict()
        n = dict()

        # Call submodule functions
        n["markdup"] = self.parse_sambamba_markdup()

        if n["markdup"] > 0:
            log.info(f"Found {n['markdup']} sambamba markdup reports")

        # Exit if we didn't find anything
        if sum(n.values()) == 0:
            raise ModuleNoSamplesFound

        # Add to the General Stats table (has to be called once per MultiQC module)
        self.general_stats_addcols(self.general_stats_data, self.general_stats_headers)



""" MultiQC module to parse output from Lima """

import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialse the parent object
        super(MultiqcModule, self).__init__(
            name="Lima",
            anchor="lima",
            href="https://github.com/PacificBiosciences/barcoding",
            info=" is used to demultiplex PacBio single-molecule sequencing reads.",
            # No publication / DOI // doi=
        )

        # To store the summary data
        self.lima_summary = dict()
        self.lima_counts = dict()

        # Parse the output files
        self.parse_summary_files()
        self.parse_counts_files()

        # Remove filtered samples
        self.lima_summary = self.ignore_samples(self.lima_summary)
        self.lima_counts = self.ignore_samples(self.lima_counts)

        # Write the data files to disk
        if not self.lima_summary and not self.lima_counts:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(self.lima_summary)} reports")
        log.info(f"Found {len(self.lima_counts)} samples")

        if self.lima_summary:
            self.write_data_file(self.lima_summary, "multiqc_lima_summary")
        if self.lima_counts:
            self.write_data_file(self.lima_counts, "multiqc_lima_counts")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Add a graph of all filtered ZMWs
        self.plot_filter_data()

        # Here, we make a difference between samples that have been renamed,
        # and samples that have their name derived from lima itself. Since lima
        # uses barcode1--barcode2 as sample names, these will never match the
        # other samples in the MultiQC report.
        #
        # Therefore, we do two different things here.
        # 1. All samples that have been renamed will be added to the general
        #       statistics table.
        # 2. All samples that have not been renamed will be added to their own
        #       table in the Lima section of the report.
        lima_renamed_count = dict()
        lima_original_count = dict()

        for sample in self.lima_counts:
            if sample in config.sample_names_replace.values():
                lima_renamed_count[sample] = self.lima_counts[sample]
            else:
                lima_original_count[sample] = self.lima_counts[sample]

        # Get the headers and tconfig for the table
        headers, tconfig = self.make_headers_config()

        # Add a graph for the data values in the counts file, for the samples
        # that haven't been renamed
        if lima_original_count:
            self.make_counts_table(lima_original_count, headers, tconfig)

        # Add renamed samples to the general statistics table, since we assume
        # they are named consistenly now
        if lima_renamed_count:
            self.add_general_stats(lima_renamed_count, headers)

    def parse_summary_files(self):
        for f in self.find_log_files("lima/summary", filehandles=True):
            data = parse_PacBio_log(f["f"])
            if len(data) > 0:
                if f["s_name"] in self.lima_summary:
                    log.debug(f"Duplicate summary sample name found! Overwriting: {f['s_name']}")
                self.lima_summary[f["s_name"]] = data
                self.add_data_source(f)

    def parse_counts_files(self):
        for f in self.find_log_files("lima/counts", filehandles=True):
            data = self.parse_lima_counts(f["f"], f)
            # Check for duplicate samples
            for sample in data:
                if sample in self.lima_counts:
                    log.debug(f"Duplicate counts sample name found! Overwriting: {sample}")
            # After warning the user, overwrite the samples
            self.lima_counts.update(data)
            self.add_data_source(f)

    def parse_lima_counts(self, file_content, f):
        """Parse lima counts file"""

        # The first line is the header
        header = next(file_content).strip().split()

        # A dictionary to store the results
        lima_counts = dict()
        for line in file_content:
            spline = line.strip().split()
            data = {field: value for field, value in zip(header, spline)}

            first_barcode = data["IdxFirstNamed"]
            second_barcode = data["IdxCombinedNamed"]
            # The format barcode1--barcode2 is also used in
            sample = self.clean_s_name(f"{first_barcode}--{second_barcode}", f)
            counts = data["Counts"]
            mean_score = data["MeanScore"]
            lima_counts[sample] = {"Counts": int(counts), "MeanScore": float(mean_score)}

        return lima_counts

    def plot_filter_data(self):
        # First, we gather all filter results for each lima summary
        all_filters = set()
        for data in self.lima_summary.values():
            for reason in self.filter_and_pass(data):
                all_filters.add(reason)
        all_filters = list(all_filters)
        # Move passing reads to the start of the list
        all_filters.insert(0, all_filters.pop(all_filters.index("ZMWs above all thresholds")))

        # Flatten the data into a structure that can be plotted
        plot_data = dict()
        for s_name, data in self.lima_summary.items():
            plot_data[s_name] = dict()
            for reason in all_filters:
                plot_data[s_name][reason] = dict()
                # This is where the filter reasons are stored
                zmw_marginals = data["ZMWs below any threshold"]["ZMW marginals"]
                for filter_reason in zmw_marginals:
                    if filter_reason == reason:
                        plot_data[s_name][reason] = zmw_marginals[filter_reason]["count"]
                        break
                # If we didn't find filter reason for this filename, set it to zero
                else:
                    plot_data[s_name][reason] = 0
                # Unless 'reason' is actually the special case of the number of unfiltered reads
                if reason == "ZMWs above all thresholds":
                    plot_data[s_name][reason] = data["ZMWs above all thresholds"]["count"]

        # Plot configuration
        config = {
            "id": "lima-filter-graph",
            "title": "Lima: ZMW filtering results",
            "ylab": "Number of ZMWs",
            "xlab": "Lima summary file",
        }

        self.add_section(
            name="ZMW filtering",
            anchor="lima-filter",
            description=("The number of ZMWs that failed or passed all Lima filters"),
            helptext="""
                The number of ZMWs that passed all filters is shown as **ZMWs above all thresholds**,
                all other categories that are shown in the graph represent the number of ZMWs that
                were dropped for the specified reason.
            """,
            plot=bargraph.plot(plot_data, all_filters, config),
        )

    def filter_and_pass(self, data):
        """Get the reasons why each ZMW was filtered"""
        reasons = dict()

        # Add why ZMWs were filtered
        filter_data = data["ZMWs below any threshold"]["ZMW marginals"]
        for reason in filter_data:
            reasons[reason] = filter_data[reason]["count"]
        # Add the ZMWs that passed
        reasons["ZMWs above all thresholds"] = data["ZMWs above all thresholds"]["count"]

        return reasons

    def make_headers_config(self):
        """Prepare the headers and config for the lima counts table"""
        tconfig = {
            "id": "multiqc_lima_counts",
            "namespace": "Lima",
            "title": "Lima: Number of Reads",
            "anchor": "multiqc_lima_counts",
            "ylab": "# Reads",
        }

        headers = {
            "Counts": {
                "title": f"Read Count ({config.long_read_count_prefix})",
                "description": f"Number of reads for each sample or barcode pair ({config.long_read_count_desc})",
                "modify": lambda x: x * config.long_read_count_multiplier,
                "shared_key": "long_read_counts",
                "format": "{:,.2f}",
                "scale": "PuBuGn",
            },
            "MeanScore": {
                "title": "Quality Score",
                "description": "The mean quality score of the reads for each sample or barcode pair",
                "scale": "Spectral",
            },
        }
        return headers, tconfig

    def make_counts_table(self, counts, headers, tconfig):
        self.add_section(
            name="Per sample count data",
            anchor="multiqc_lima_count",
            description="""
                Per sample or per barcode statistics from Lima.
                For instructions on how to display sample names instead of `barcode--barcode` pairs,
                please see the [MultiQC Lima documentation](https://multiqc.info/docs/#lima).
            """,
            plot=table.plot(counts, headers, tconfig),
        )

    def add_general_stats(self, counts, headers):
        """Add (renamed) samples to the general statistics table"""
        self.general_stats_addcols(counts, headers)


def parse_PacBio_log(file_content):
    """Parse PacBio log file"""
    data = dict()
    # This is a local dictionary to store which annotations belong to which
    # result dictionary. This will be used to structure the output, but will
    # not be part of the output itself
    annotations = dict()
    current_annotation = None

    for line in file_content:
        # Get rid of trailing newlines
        line = line.strip("\n")
        # Did we enter a new section with annotations for an earlier result?
        # If so, we will only add an empty dictionary with the correct name
        # These field are of the format "something something for (A):"
        section_header_pattern = " for [(][A-Z][)][:]$"
        if re.search(section_header_pattern, line):
            linedata = parse_line(line)
            ann = linedata["annotation"]
            # Cut off the ' for (B):' part
            name = line[:-9]
            # We make a new heading with the current name under the data that
            # matches the current annotation
            current_annotation = dict()
            # We keep the dictonary accessible under 'current_annotation',
            # so we can keep adding new data to it without having to keep track
            # of where it belongs
            annotations[ann][name] = current_annotation
            continue

        linedata = parse_line(line)

        # If we got no data, we reached the end of the annotated section
        if not linedata:
            current_annotation = None
            continue

        # Lets get the name of the data
        name = linedata.pop("name")
        # If we are in an annotated section, we add the data to the current
        # annotation
        if current_annotation is not None:
            current_annotation[name] = linedata
        # Otherwise, we add the newfound annotation to the dictionary in case
        # we find a corresponding section later on.
        # The data from the current line we add directly to the output data
        else:
            annotation = linedata.pop("annotation")
            annotations[annotation] = linedata
            data[name] = linedata

    return data


def parse_line(line):
    """Parse a line from the Lima log file"""
    data = dict()

    # If we got an empty line to parse
    if not line.strip():
        return data

    # Split the line on the colon character
    key, values = line.strip().split(":")

    # The key can have multiple parts
    keys = key.strip().split()

    # Does the key have an annotation (A), (B) etc
    if re.fullmatch("[(][A-Z][)]", keys[-1]):
        # We store the annotation without the bracets
        data["annotation"] = keys[-1][1:-1]
        # And we add the rest of the key as name
        data["name"] = " ".join(keys[:-1])
    # Otherwise, we just store the name
    else:
        data["name"] = " ".join(keys)

    # Parsing the values
    values = values.strip().split()
    # Are there are no values we are done
    if not values:
        return data

    # If there is a single value
    if len(values) == 1:
        value = values.pop()
        # Is the value a percentage
        if value.endswith("%"):
            data["percentage"] = float(value[:-1])
        # Otherwise, it is an integer
        else:
            data["count"] = int(value)
    elif len(values) == 2:
        # If there are multiple values, they are in the format
        # count (percentage%)
        count = values[0]
        percentage = values[1]

        # The percentage should be in the format:
        # (12.34%) or (100%) or (0%) or (-nan%)
        # So we can remove the bracets first
        percentage = percentage[1:-1]
        # Then we make sure it is one of the supported patterns
        assert re.fullmatch(r"(\d+\.\d+%|\d+%|-nan%)", percentage)

        # If the percentage is this weird nan, set it to 0
        if percentage == "-nan%":
            data["percentage"] = 0.0
        else:  # Otherwise, we cut of the % sign and convert to float
            data["percentage"] = float(percentage[:-1])
        # Add the count as an integer
        data["count"] = int(count)

    return data



# -*- coding: utf-8 -*-

""" MultiQC submodule to parse output from GATK tool AnalyzeSaturationMutagenesis """

import logging
from collections import defaultdict

from multiqc import config
from multiqc.plots import bargraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class AnalyzeSaturationMutagenesisMixin:
    """Mixin class to add GATK AnalyzeSaturationMutagenesis reports to MultiQC."""

    def parse_gatk_analyze_saturation_mutagenesis(self):
        """Find GATK AnalyzeSaturationMutagenesis logs and parse their data"""

        self.gatk_analyze_saturation_mutagenesis = {}

        for file_handle in self.find_log_files("gatk/analyze_saturation_mutagenesis", filehandles=True):
            parsed_data = self.parse_read_counts_file(file_handle["f"])
            if len(parsed_data) > 1:
                if file_handle["s_name"] in self.gatk_analyze_saturation_mutagenesis:
                    log.debug(f"Duplicate sample name found! Overwriting: {file_handle['s_name']}")
                self.add_data_source(file_handle, section="analyze_saturation_mutagenesis")
                self.gatk_analyze_saturation_mutagenesis[file_handle["s_name"]] = parsed_data

        # Filter to strip out ignored sample names
        self.gatk_analyze_saturation_mutagenesis = self.ignore_samples(self.gatk_analyze_saturation_mutagenesis)

        n_reports_found = len(self.gatk_analyze_saturation_mutagenesis)
        if n_reports_found == 0:
            return 0

        log.info("Found %d AnalyzeSaturationMutagenesis reports", n_reports_found)

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file (restructure first)
        self.write_data_file(self.gatk_analyze_saturation_mutagenesis, "multiqc_gatk_analyze_saturation_mutagenesis")

        self.gatk_analyze_saturation_mutagenesis_table(self.gatk_analyze_saturation_mutagenesis)
        # Add plots
        self.gatk_analyze_saturation_mutagenesis_plot_reads(self.gatk_analyze_saturation_mutagenesis)
        self.gatk_analyze_saturation_mutagenesis_plot_base_calls(self.gatk_analyze_saturation_mutagenesis)

        return n_reports_found

    def parse_read_counts_file(self, file_handle):
        """Parse a readCounts output file from GATK AnalyzeSaturationMutagenesis
        These files are tab delimited, with some hierarchical structuring.

        This does not parse percentages, since those are readily calculated
        and MultiQC automatically generates them for plotting."""

        label_to_key = {
            "Total Reads:": "total_reads",
            ">Unmapped Reads:": "unmapped_reads",
            ">LowQ Reads:": "lowq_reads",
            ">Evaluable Reads:": "evaluable_reads",
            ">>Reads in disjoint pairs evaluated separately:": "disjoint_pairs",
            ">>Reads in overlapping pairs evaluated together:": "overlapping_pairs",
            ">>>Wild type:": "wt_reads",
            ">>>Called variants:": "called_variants",
            ">>>Mate ignored:": "mate_ignored",
            ">>>Inconsistent pair:": "inconsistent",
            ">>>Low quality variation:": "low_quality_variation",
            ">>>Insufficient flank:": "insufficient_flank",
            "Total base calls:": "total_base_calls",
            ">Base calls evaluated for variants:": "evaluated_base_calls",
            ">Base calls unevaluated:": "unevaluated_base_calls",
        }

        data = defaultdict(int)

        # For keeping track of whether we are in disjoint or overlapping reads
        # Disjoint are first, so set to True initially
        disjoint_reads = True

        for line in file_handle:
            fields = line.split("\t")
            label = fields[0]
            key = label_to_key[label]
            if key == "disjoint_pairs":
                disjoint_reads = True
            elif key == "overlapping_pairs":
                disjoint_reads = False
            elif label.startswith(">>>"):
                key = key + "_" + ("disjoint" if disjoint_reads else "overlapping")
            data[key] = int(fields[1])

        # Create some summary fields
        data["wt_total_reads"] = data["wt_reads_disjoint"] + data["wt_reads_overlapping"]
        data["variants_total_reads"] = data["called_variants_disjoint"] + data["called_variants_overlapping"]
        data["filtered_reads"] = sum(
            [
                data["lowq_reads"],
                data["mate_ignored_disjoint"],
                data["inconsistent_overlapping"],
                data["low_quality_variation_disjoint"],
                data["low_quality_variation_overlapping"],
                data["insufficient_flank_disjoint"],
                data["insufficient_flank_overlapping"],
            ]
        )
        return data

    def gatk_analyze_saturation_mutagenesis_plot_reads(self, data):
        """Make the plot for GATK AnalyzeSaturationMutagenesis read counts and add the section."""
        cats = {
            "filtered_reads": {"name": "Filtered reads"},
            "wt_total_reads": {"name": "WT reads"},
            "variants_total_reads": {"name": "Variant reads"},
        }

        pconfig = {
            "id": "gatk_ASM_reads_plot",
            "title": "GATK AnalyzeSaturationMutagenesis: Read counts",
            "ylab": "Number of reads",
            "cpswitch_counts_label": "Counts",
        }

        self.add_section(
            name="Read counts",
            anchor="gatk-asm-read-counts",
            description="Read counts and read fate. Filtered reads include unmapped, low quality, and other pathologies.",
            helptext="""Reads can be filtered by GATK AnalyzeSaturationMutagenesis for a number of reasons, including low quality, insufficient flank, and other pathologies.
            This plot shows the number of reads that were mapped to WT, called as variants, and the number that were filtered.""",
            plot=bargraph.plot(data, cats, pconfig),
        )

    def gatk_analyze_saturation_mutagenesis_plot_base_calls(self, data):
        """Make the plot for GATK AnalyzeSaturationMutagenesis base calls and add the section."""
        cats = {
            "evaluated_base_calls": {"name": "Base calls evaluated for variants"},
            "unevaluated_base_calls": {"name": "Base calls not evaluated for variants"},
        }

        pconfig = {
            "id": "gatk_ASM_base_calls_plot",
            "title": "GATK AnalyzeSaturationMutagenesis: Base calls",
            "ylab": "Number of bases",
            "cpswitch_counts_label": "Counts",
        }

        self.add_section(
            name="Base calls",
            anchor="gatk-asm-bases",
            description="Base calls evaluated for variants and base calls not evaluated for variants.",
            helptext="""Bases can be filtered by GATK AnalyzeSaturationMutagenesis for a number of reasons, including low quality, insufficient flank, and other pathologies.
             This plot shows the number of base calls that were evaluated for variants and the number of base calls that were not evaluated for variants.""",
            plot=bargraph.plot(data, cats, pconfig),
        )

    def gatk_analyze_saturation_mutagenesis_table(self, data):
        """Make the table for GATK AnalyzeSaturationMutagenesis and add the section."""
        asm_headers = {
            "total_reads": {
                "title": f"Total reads ({config.read_count_prefix})",
                "description": f"Total reads in sample ({config.read_count_desc})",
                "min": 0,
                "scale": "Greys",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "wt_total_reads": {
                "title": f"WT reads ({config.read_count_prefix})",
                "description": f"Total evaluated reads mapped to WT ({config.read_count_desc})",
                "min": 0,
                "scale": "Blues",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "variants_total_reads": {
                "title": f"Variant reads ({config.read_count_prefix})",
                "description": f"Reads with a variant called ({config.read_count_desc})",
                "min": 0,
                "scale": "Greens",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "filtered_reads": {
                "title": f"Filtered reads ({config.read_count_prefix})",
                "description": f"Reads filtered from sample ({config.read_count_desc})",
                "min": 0,
                "scale": "Reds-rev",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "unmapped_reads": {
                "title": f"Unmapped ({config.read_count_prefix})",
                "description": f"Unmapped reads in sample ({config.read_count_desc})",
                "min": 0,
                "scale": "Purples-rev",
                "shared_key": "read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "lowq_reads": {
                "title": f"LowQ ({config.read_count_prefix})",
                "description": f"Low quality reads in sample ({config.read_count_desc})",
                "min": 0,
                "scale": "Blues-rev",
                "shared_key": "read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "evaluable_reads": {
                "title": f"Evaluable ({config.read_count_prefix})",
                "description": f"Evaluable reads in sample ({config.read_count_desc})",
                "min": 0,
                "scale": "Greys",
                "shared_key": "read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "disjoint_pairs": {
                "title": f"Disjoint ({config.read_count_prefix})",
                "description": f"Reads from disjoint (non-overlapping) paired-end reads in sample ({config.read_count_desc})",
                "min": 0,
                "scale": "Oranges",
                "shared_key": "read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "wt_reads_disjoint": {
                "title": f"WT (disjoint) ({config.read_count_prefix})",
                "description": f"WT reads called from disjoint pairs ({config.read_count_desc})",
                "min": 0,
                "scale": "Purples",
                "shared_key": "disjoint_read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "called_variants_disjoint": {
                "title": f"Called variants (disjoint) ({config.read_count_prefix})",
                "description": f"Reads with variants called from disjoint pairs ({config.read_count_desc})",
                "min": 0,
                "scale": "Blues",
                "shared_key": "disjoint_read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "mate_ignored_disjoint": {
                "title": f"Mateless ({config.read_count_prefix})",
                "description": f"Reads with ignored mates called from disjoint pairs ({config.read_count_desc})",
                "min": 0,
                "scale": "Greens-rev",
                "shared_key": "disjoint_read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "low_quality_variation_disjoint": {
                "title": f"LowQ (disjoint) ({config.read_count_prefix})",
                "description": f"Reads with low quality variation from disjoint pairs ({config.read_count_desc})",
                "min": 0,
                "scale": "Reds-rev",
                "shared_key": "disjoint_read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "insufficient_flank_disjoint": {
                "title": f"No flank (disjoint) ({config.read_count_prefix})",
                "description": f"Reads with insufficient flank from disjoint pairs ({config.read_count_desc})",
                "min": 0,
                "scale": "Blues-rev",
                "shared_key": "disjoint_read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "overlapping_pairs": {
                "title": f"Overlapping ({config.read_count_prefix})",
                "description": f"Reads from overlapping paired-end reads in sample ({config.read_count_desc})",
                "min": 0,
                "scale": "Oranges",
                "shared_key": "overlapping_read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "wt_reads_overlapping": {
                "title": f"WT (overlapping) ({config.read_count_prefix})",
                "description": f"WT reads called from overlapping pairs ({config.read_count_desc})",
                "min": 0,
                "scale": "Purples",
                "shared_key": "overlapping_read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "called_variants_overlapping": {
                "title": f"Called variants (overlapping) ({config.read_count_prefix})",
                "description": f"Reads with variants called from overlapping pairs ({config.read_count_desc})",
                "min": 0,
                "scale": "Blues",
                "shared_key": "overlapping_read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "inconsistent_overlapping": {
                "title": f"Inconsistent (overlapping) ({config.read_count_prefix})",
                "description": f"Reads with inconsistent pairs from overlapping pairs ({config.read_count_desc})",
                "min": 0,
                "scale": "Greens-rev",
                "shared_key": "overlapping_read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "low_quality_variation_overlapping": {
                "title": f"Low quality reads (overlapping) ({config.read_count_prefix})",
                "description": f"Reads with low quality variation from overlapping pairs ({config.read_count_desc})",
                "min": 0,
                "scale": "Reds-rev",
                "shared_key": "overlapping_read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "insufficient_flank_overlapping": {
                "title": f"No flank (overlapping) ({config.read_count_prefix})",
                "description": f"Reads with insufficient flank from overlapping pairs ({config.read_count_desc})",
                "min": 0,
                "scale": "Blues-rev",
                "shared_key": "overlapping_read_count",
                "hidden": True,
                "modify": lambda x: x * config.read_count_multiplier,
                "namespace": "GATK",
            },
            "total_base_calls": {
                "title": f"Total bases ({config.base_count_prefix})",
                "description": f"Total base calls in sample ({config.base_count_desc})",
                "min": 0,
                "scale": "Greys",
                "shared_key": "base_calls",
                "hidden": True,
                "modify": lambda x: x * config.base_count_multiplier,
                "namespace": "GATK",
            },
            "evaluated_base_calls": {
                "title": f"Evaluated bases ({config.base_count_prefix})",
                "description": f"Evaluated base calls in sample ({config.base_count_desc})",
                "min": 0,
                "scale": "Blues",
                "hidden": True,
                "shared_key": "base_calls",
                "modify": lambda x: x * config.base_count_multiplier,
                "namespace": "GATK",
            },
            "unevaluated_base_calls": {
                "title": f"Unevaluated bases ({config.base_count_prefix})",
                "description": f"Unevaluated base calls in sample ({config.base_count_desc})",
                "min": 0,
                "scale": "Reds-rev",
                "hidden": True,
                "shared_key": "base_calls",
                "modify": lambda x: x * config.base_count_multiplier,
                "namespace": "GATK",
            },
        }

        # Add module specific prefix to all keys to be safe
        prefix = "gatk_ask_"
        asm_headers = {f"{prefix}{k}": v for k, v in asm_headers.items()}
        data = {sn: {f"{prefix}{k}": v for k, v in d.items()} for sn, d in data.items()}

        pconfig = {"id": f"{prefix}stats", "namespace": "GATK", "table_title": "GATK ASM counts"}

        self.add_section(
            name="GATK ASM counts",
            anchor="gatk-asm-stats",
            description="Per-sample read count and base call fates from GATK AnalyzeSaturationMutagenesis.",
            helptext="""
            This table shows the distribution of calls (for reads or for bases) across all samples.
            Reads are categorized as WT, a variant, or filtered. Bases can be either evaluated or unevaluated, corresponding to the reads they come from.

            Reads are filtered for the following reasons:

            * Unmapped: the map quality is below the minimum MapQ (default = 4)'
            * Low quality reads: Reads are trimmed for quality > minQ (default 30) before calling variants. If the final size is less than the minimum length (default 15), or if the remaining segment does not cover the ORF, the read is filtered.

            Paired reads are also split into overlapping and disjoint sets, with further filters for both.

            * Inconsistent: If overlapping reads disagree on the called variant, the read is filtered.
            * Ignored mate: If the pairs are disjoint, the first read of the pair is used for analysis, and the second is ignored.
            * Low quality variation: If the variant includes ambiguous bases (not A, C, G, or T, or -), the read is filtered.
            * Insufficient flank: If the variant does not include a certain number of WT bases (default 2) flanking the variant, the read is filtered.
            """,
            plot=table.plot(data, asm_headers, pconfig),
        )



""" MultiQC submodule to parse output from Picard AlignmentSummaryMetrics """

import logging
from collections import OrderedDict

from multiqc.modules.picard import util
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(module):
    """Find Picard AlignmentSummaryMetrics reports and parse their data"""

    data_by_sample = dict()

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/alignment_metrics", filehandles=True):
        # Sample name from input file name by default.
        s_name = f["s_name"]
        keys = None

        for line in f["f"]:
            maybe_s_name = util.extract_sample_name(
                module,
                line,
                f,
                picard_tool="CollectAlignmentSummaryMetrics",
                sentieon_algo="AlignmentStat",
            )
            if maybe_s_name:
                s_name = maybe_s_name
                keys = None

            if s_name is None:
                continue

            if util.is_line_right_before_table(
                line, picard_class="AlignmentSummaryMetrics", sentieon_algo="AlignmentStat"
            ):
                if s_name in data_by_sample:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: " f"{s_name}")
                data_by_sample[s_name] = dict()
                module.add_data_source(f, s_name, section="AlignmentSummaryMetrics")
                keys = f["f"].readline().strip("\n").split("\t")

            elif keys:
                vals = line.strip("\n").split("\t")
                if len(vals) != len(keys):
                    keys = None
                    continue

                # Ignore the FIRST_OF_PAIR / SECOND_OF_PAIR data to simplify things
                if vals[0] == "PAIR" or vals[0] == "UNPAIRED":
                    for k, v in zip(keys, vals):
                        try:
                            v = float(v)
                        except ValueError:
                            pass
                        data_by_sample[s_name][k] = v

    # Filter to strip out ignored sample names
    data_by_sample = module.ignore_samples(data_by_sample)
    if len(data_by_sample) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    module.write_data_file(data_by_sample, f"multiqc_{module.anchor}_AlignmentSummaryMetrics")

    # Add to general stats table
    module.general_stats_headers["PCT_PF_READS_ALIGNED"] = {
        "title": "% Aligned",
        "description": "Percent of aligned reads",
        "max": 100,
        "min": 0,
        "suffix": "%",
        "format": "{:,.0f}",
        "scale": "RdYlGn",
        "modify": lambda x: util.multiply_hundred(x),
    }
    for s_name in data_by_sample:
        if s_name not in module.general_stats_data:
            module.general_stats_data[s_name] = dict()
        module.general_stats_data[s_name].update(data_by_sample[s_name])

    # Make the bar plot of alignment read count + # aligned bases
    pdata = dict()
    for s_name in data_by_sample.keys():
        pdata[s_name] = dict()
        # Picard reports both reads for PE data. Divide it by two as most people will
        # expect # clusters
        if data_by_sample[s_name]["CATEGORY"] == "PAIR":
            pdata[s_name]["total_reads"] = data_by_sample[s_name]["TOTAL_READS"] / 2
            pdata[s_name]["aligned_reads"] = data_by_sample[s_name]["PF_READS_ALIGNED"] / 2
        else:
            pdata[s_name]["total_reads"] = data_by_sample[s_name]["TOTAL_READS"]
            pdata[s_name]["aligned_reads"] = data_by_sample[s_name]["PF_READS_ALIGNED"]
        pdata[s_name]["unaligned_reads"] = pdata[s_name]["total_reads"] - pdata[s_name]["aligned_reads"]

    keys = [OrderedDict(), OrderedDict()]
    keys[0]["aligned_reads"] = {"name": "Aligned Reads"}
    keys[0]["unaligned_reads"] = {"name": "Unaligned Reads"}
    keys[1]["PF_ALIGNED_BASES"] = {"name": "Aligned Bases"}

    # Config for the plot
    pconfig = {
        "id": f"{module.anchor}_alignment_summary",
        "title": f"{module.name}: Alignment Summary",
        "ylab": "# Reads",
        "data_labels": [
            {
                "name": "Aligned Reads",
                "ylab": "# Reads",
                "cpswitch_counts_label": "Number of Reads",
            },
            {
                "name": "Aligned Bases",
                "ylab": "# Bases",
                "cpswitch_counts_label": "Number of Bases",
            },
        ],
    }

    # The different data sets we want to plot
    module.add_section(
        name="Alignment Summary",
        anchor=f"{module.anchor}-alignmentsummary",
        description=f"Please note that {module.name}'s read counts are divided by two "
        f"for paired-end data. Total bases (including unaligned) is not "
        f"provided.",
        plot=bargraph.plot([pdata, data_by_sample], keys, pconfig),
    )

    # Make a bar plot of mean read length
    keys = {"MEAN_READ_LENGTH": {"name": "Mean Read Length"}}
    pconfig = {
        "id": f"{module.anchor}_alignment_readlength_plot",
        "title": f"{module.name}: Mean Read Length",
        "ylab": "Base pairs",
        "cpswitch": False,
    }

    # The different data sets we want to plot
    module.add_section(
        name="Mean read length",
        anchor=f"{module.anchor}_alignment_readlength",
        description="The mean read length of the set of reads examined.",
        plot=bargraph.plot(data_by_sample, keys, pconfig),
    )

    # Return the number of detected samples to the parent module
    return len(data_by_sample)



""" MultiQC module to parse output from the Seqera Platform CLI """

import datetime as dt
import json
import logging
import os
import re
import tarfile
from collections import defaultdict

import humanize

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph
from multiqc.utils import mqc_colour

log = logging.getLogger(__name__)


def _read_json_from_tar_gz(tar_file, fname):
    try:
        fh = tar_file.extractfile(tar_file.getmember(fname))
        contents = fh.read()
    except Exception as e:
        log.warning(f"Could not extract file {fname} from archive {tar_file}: {e}")
        return {}
    try:
        data = json.loads(contents)
    except Exception as e:
        log.warning(f"Could parse JSON from {fname} in {tar_file}: {e}")
        return {}
    return data


class MultiqcModule(BaseMultiqcModule):
    """
    Seqera Platform CLI module for MultiQC. Reports stats from log dumps
    usually written in a form of a tar-gz archive, but also their uncompressed
    versions (that is, workflow.json and workflow-load.json files).
    To allow reading the tar-gz archives, run with `ignore_images: false`
    in the config, e.g.:
    ```
    multiqc . --cl-config 'ignore_images: false'
    ```
    """

    def __init__(self):
        super(MultiqcModule, self).__init__(
            name="Seqera Platform CLI",
            anchor="seqera_cli",
            href="https://github.com/seqeralabs/tower-cli",
            info="reports statistics generated by the Seqera Platform CLI.",
            # doi="",  # No DOI for this tool
        )

        data_by_run = defaultdict(dict)

        # Parsing the tar-gz dump
        for f in self.find_log_files("seqera_cli/run_dump", filecontents=False):
            with tarfile.open(os.path.join(f["root"], f["fn"])) as tar_file:
                if "workflow.json" not in tar_file.getnames():
                    continue
                d = _read_json_from_tar_gz(tar_file, "workflow.json")

                # Check other files that sit next to workflow.json
                if "workflow-load.json" in tar_file.getnames():
                    d.update(_read_json_from_tar_gz(tar_file, "workflow-load.json"))

                if "service-info.json" in tar_file.getnames():
                    d.update(_read_json_from_tar_gz(tar_file, "service-info.json"))

                if "workflow-metadata.json" in tar_file.getnames():
                    d.update(_read_json_from_tar_gz(tar_file, "workflow-metadata.json"))

                d = self._parse_data(d)
                if d:
                    self.add_data_source(f, s_name=d["id_repository"])
                    data_by_run[d["id_repository"]].update(d)

        # Parsing the json files directly
        for f in self.find_log_files("seqera_cli/json"):
            d = json.loads(f["f"])
            if not d:
                continue

            # Check other files that sit next to workflow.json
            load_path = os.path.join(f["root"], "workflow-load.json")
            if os.path.isfile(load_path):
                with open(load_path) as fh:
                    d.update(json.load(fh))

            service_info_path = os.path.join(f["root"], "service-info.json")
            if os.path.isfile(service_info_path):
                with open(service_info_path) as fh:
                    d.update(json.load(fh))

            workflow_metadata_path = os.path.join(f["root"], "workflow-metadata.json")
            if os.path.isfile(workflow_metadata_path):
                with open(workflow_metadata_path) as fh:
                    d.update(json.load(fh))

            d = self._parse_data(d)
            if d:
                self.add_data_source(f, s_name=d["id_repository"])
                data_by_run[d["id_repository"]].update(d)
                if os.path.isfile(load_path):
                    self.add_data_source(source=load_path, s_name=d["id_repository"])
                if os.path.isfile(service_info_path):
                    self.add_data_source(source=service_info_path, s_name=d["id_repository"])

        # Figure out the org/workspace and save as a separate field
        # Needed so that we can have it as a separate column in the table
        for d in data_by_run.values():
            run_url_re = re.compile(r"\/orgs\/([^\/]+)\/workspaces\/([^\/]+)\/watch\/([^\/]+)\/?$")
            if d.get("runUrl"):
                m = run_url_re.search(d["runUrl"])
                if m:
                    org, workspace, run = m.groups()
                    d["org"] = org
                    d["workspace"] = workspace
                    d["org_workspace"] = f"{org}/{workspace}"

        # Filter to strip out ignored sample names
        data_by_run = self.ignore_samples(data_by_run)
        if len(data_by_run) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(data_by_run)} reports")

        # Write parsed report data to a file
        self.write_data_file(data_by_run, "multiqc_seqera_cli")

        self._table(data_by_run)

        self._plots(data_by_run)

    def _parse_data(self, d):
        keys = [
            # workflow.json
            "id",
            "runUrl",
            "repository",
            "start",
            "complete",
            "revision",
            "nextflow",
            # workflow-load.json
            "cpuEfficiency",
            "memoryEfficiency",
            "cpuTime",
            "readBytes",
            "writeBytes",
            "cost",
            "pending",
            "submitted",
            "running",
            "succeeded",
            "failed",
            "cached",
            # service-info.json
            "version",
        ]
        d = {k: d.get(k) for k in keys if k in d}

        if not d.get("id") or not d.get("repository"):
            return None

        repo = d["repository"]
        repo = repo.replace("https://", "").replace("http://", "").replace("github.com/", "")
        d["id_repository"] = f"{repo}_{d['id']}"

        # "start" and "complete" are time stamps like time stamps like 2023-10-22T14:39:01Z
        # parse them with a library, take the difference "complete" - "start" to get the
        # wall time, and convert the wall time it to a human-readable format.
        if "start" in d and "complete" in d and d["complete"] is not None:
            start = dt.datetime.strptime(d["start"], "%Y-%m-%dT%H:%M:%SZ")
            complete = dt.datetime.strptime(d["complete"], "%Y-%m-%dT%H:%M:%SZ")
            wall_time = complete - start
            d["wallTime"] = wall_time.total_seconds()
            d["start"] = start.timestamp()
            d["complete"] = complete.timestamp()

        version = d.get("version")
        if version:
            d["seqeraVersion"] = version
            self.add_software_version(version, sample=d["id_repository"], software_name="Seqera Platform")

        nextflow_version = d.get("nextflow", {}).get("version")
        if nextflow_version:
            d["nextflowVersion"] = nextflow_version
            self.add_software_version(nextflow_version, sample=d["id_repository"], software_name="Nextflow")

        return d

    def _table(self, data_by_run):
        """
        Add the main table with run statistics.
        """
        # Collecting categorical values into distinct lists that we want to color code
        # with badges and backgrounds:
        seqera_versions = list(set(d.get("seqeraVersion") for d in data_by_run.values()))
        nextflow_versions = list(set(d.get("nextflowVersion") for d in data_by_run.values()))
        scale = mqc_colour.mqc_colour_scale("Dark2")
        version_colors = [
            {v: scale.get_colour(i, lighten=0.5)} for i, v in enumerate(seqera_versions + nextflow_versions)
        ]
        repositories = list(set(d.get("repository") for d in data_by_run.values()))

        def format_run_url(x):
            runUrl_re = re.compile(r"\/orgs\/([^\/]+)\/workspaces\/([^\/]+)\/watch\/([^\/]+)\/?$")
            if x:
                m = runUrl_re.search(x)
                if m:
                    org, workspace, run = m.groups()
                    return f'<a href="{x}" target="_blank">{run}</a>'
            return str(x)

        headers = {
            "runUrl": {
                "title": "Run ID",
                "description": "Workflow run ID",
                "scale": False,
                "format": format_run_url,
            },
            "org_workspace": {
                "title": "Workspace",
                "description": "Organisation and workspace",
                "scale": False,
            },
            "repository": {
                "title": "Repository",
                "description": "Name of the repository",
                "scale": "Accent",
                "modify": lambda x: repositories.index(x),
                "format": lambda x: f'<a href="{repositories[x]}" style="white-space: nowrap;">{repositories[x].replace("https://", "").replace("http://", "").replace("github.com/", "")}</a>',
            },
            "revision": {
                "title": "Version",
                "description": "Pipeline version",
                "scale": False,
            },
            "start": {
                "title": "Start",
                "description": "Start time of the workflow",
                "hidden": True,
                "format": lambda x: humanize.naturaltime(dt.datetime.fromtimestamp(x)),
            },
            "complete": {
                "title": "Complete",
                "description": "End time of the workflow",
                "hidden": True,
                "format": lambda x: humanize.naturaltime(dt.datetime.fromtimestamp(x)),
            },
            "wallTime": {
                "title": "Wall time",
                "description": "Duration of the workflow",
                "format": lambda x: str(dt.timedelta(seconds=x)),
                "scale": "BuPu",
            },
            "cpuTime": {
                "title": "CPU time",
                "description": "Total CPU time used by the workflow",
                "modify": lambda x: x // 1000 / 60 / 60,  # hours
                "suffix": " h",
                "scale": "Greys",
            },
            "cost": {
                "title": "Est. cost",
                "description": "Estimated cost of the workflow",
                "format": "${:,.2f}",
                "scale": "Reds",
            },
            "readBytes": {
                "title": "Read GB",
                "description": "Total gigabytes read by the workflow",
                "format": lambda x: humanize.naturalsize(x),
                "scale": "Blues",
            },
            "writeBytes": {
                "title": "Write GB",
                "description": "Total gigabytes written by the workflow",
                "format": lambda x: humanize.naturalsize(x),
                "scale": "Greens",
            },
            "cpuEfficiency": {
                "title": "CPU efficiency",
                "description": "Percentage of CPU time used by the workflow",
                "format": "{:,.2f}",
                "suffix": "%",
                "max": 100,
                "scale": "RdYlGn",
            },
            "memoryEfficiency": {
                "title": "Memory efficiency",
                "description": "Percentage of memory used by the workflow",
                "format": "{:,.2f}",
                "suffix": "%",
                "max": 100,
                "scale": "YlGn",
            },
            "nextflowVersion": {
                "title": "Nextflow",
                "description": "Version of Nextflow",
                "cond_formatting_colours": version_colors,
                "cond_formatting_rules": {v: [{"s_eq": v}] for v in nextflow_versions},
                "scale": False,
            },
        }
        self.general_stats_addcols(data_by_run, headers)

    def _plots(self, data_by_run):
        """
        Add bar plots
        """
        self.add_section(
            name="Workflow status",
            anchor="seqera_cli_process_statuses_section",
            plot=bargraph.plot(
                data_by_run,
                cats={
                    "pending": {"name": "Pending", "color": "#8f4199"},
                    "submitted": {"name": "Submitted", "color": "#e68642"},
                    "running": {"name": "Running", "color": "#4256e7"},
                    "cached": {"name": "Cached", "color": "#939598"},
                    "succeeded": {"name": "Succeeded", "color": "#28ae61"},
                    "failed": {"name": "Failed", "color": "#e7363e"},
                },
                pconfig={
                    "id": "seqera_cli_process_statuses_plot",
                    "title": "Seqera platform CLI: Workflow status",
                    "ylab": "Number of tasks",
                },
            ),
        )

        plot_data = dict()
        for sn, data in data_by_run.items():
            plot_data[sn] = dict()
            if "wallTime" in data:
                plot_data[sn]["wallTime"] = data["wallTime"] / 60 / 60  # hours
            if "cpuTime" in data is not None:
                plot_data[sn]["cpuTime"] = data["cpuTime"] // 1000 / 60 / 60  # hours
            if "cost" in data is not None:
                plot_data[sn]["cost"] = data["cost"]

        self.add_section(
            name="Wall time",
            anchor="seqera_cli_wall_time_section",
            plot=bargraph.plot(
                plot_data,
                {"wallTime": {"name": "Wall time"}},
                {
                    "id": "seqera_cli_wall_time_plot",
                    "title": "Seqera platform CLI: Wall time",
                    "ylab": "hours",
                    "tt_decimals": 1,
                    "tt_suffix": " h",
                    "tt_percentages": False,
                    "cpswitch": False,
                    "hide_zero_cats": False,
                },
            ),
        )
        self.add_section(
            name="CPU time",
            anchor="seqera_cli_cpu_time_section",
            plot=bargraph.plot(
                plot_data,
                {"cpuTime": {"name": "CPU time"}},
                {
                    "id": "seqera_cli_cpu_time_plot",
                    "title": "Seqera platform CLI: CPU time",
                    "ylab": "CPU hours",
                    "tt_decimals": 1,
                    "tt_suffix": " h",
                    "tt_percentages": False,
                    "cpswitch": False,
                    "hide_zero_cats": False,
                },
            ),
        )
        self.add_section(
            name="Estimated cost",
            anchor="seqera_cli_cost_section",
            plot=bargraph.plot(
                plot_data,
                {"cost": {"name": "Estimated cost"}},
                {
                    "id": "seqera_cli_cost_plot",
                    "title": "Seqera platform CLI: Estimated cost",
                    "ylab": "$",
                    "tt_decimals": 1,
                    "tt_suffix": " $",
                    "tt_percentages": False,
                    "cpswitch": False,
                    "hide_zero_cats": False,
                },
            ),
        )



""" MultiQC module to parse output from HiCUP """


import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """HiCUP module, parses log files saved by HiCUP."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="HiCUP",
            anchor="hicup",
            href="http://www.bioinformatics.babraham.ac.uk/projects/hicup/",
            info="(Hi-C User Pipeline) is a tool for mapping and performing quality control on Hi-C data.",
            doi="10.12688/f1000research.7334.1",
        )

        # Find and load any HiCUP summary reports
        self.hicup_data = dict()
        for f in self.find_log_files("hicup"):
            self.parse_hicup_logs(f)

        # Filter to strip out ignored sample names
        self.hicup_data = self.ignore_samples(self.hicup_data)

        if len(self.hicup_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.hicup_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed data to a file
        self.write_data_file(self.hicup_data, "multiqc_hicup")

        # Basic Stats Table
        self.hicup_stats_table()

        # Report sections
        self.add_section(name="Read Truncation", anchor="hicup-truncating", plot=self.hicup_truncating_chart())
        self.add_section(name="Read Mapping", anchor="hicup-mapping", plot=self.hicup_alignment_chart())

        self.add_section(name="Read Pair Filtering", anchor="hicup-filtering", plot=self.hicup_filtering_chart())

        # TODO: Is there a log file with this data for a line plot?
        # self.add_section (
        #     name = 'Di-Tag Length Distribution',
        #     anchor = 'hicup-lengths',
        #     plot = self.hicup_lengths_chart()
        # )

        self.add_section(
            name="De-Duplication &amp; Di-Tag Separation", anchor="hicup-deduplication", plot=self.hicup_dedup_chart()
        )

    def parse_hicup_logs(self, f):
        """Parse a HiCUP summary report"""
        if not f["fn"].endswith(".txt"):
            return None
        header = []
        lines = f["f"].splitlines()
        for line in lines:
            s = line.split("\t")
            if len(header) == 0:
                if s[0] != "File":
                    return None
                header = s[1:]
            else:
                s_name = self.clean_s_name(s[0], f)
                if s_name.startswith("HiCUP_output/"):
                    s_name = s_name[13:]
                parsed_data = {}
                for idx, num in enumerate(s[1:]):
                    try:
                        parsed_data[header[idx]] = float(num)
                    except ValueError:
                        parsed_data[header[idx]] = num
                parsed_data["Duplicate_Read_Pairs"] = (
                    parsed_data["Valid_Pairs"] - parsed_data["Deduplication_Read_Pairs_Uniques"]
                )
                if s_name in self.hicup_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.add_data_source(f, s_name)
                self.hicup_data[s_name] = parsed_data

    def hicup_stats_table(self):
        """Add core HiCUP stats to the general stats table"""
        headers = {
            "Percentage_Ditags_Passed_Through_HiCUP": {
                "title": "% Passed",
                "description": "Percentage Di-Tags Passed Through HiCUP",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
            },
            "Deduplication_Read_Pairs_Uniques": {
                "title": f"{config.read_count_prefix} Unique",
                "description": f"Unique Di-Tags ({config.read_count_desc})",
                "min": 0,
                "scale": "PuRd",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
            },
            "Percentage_Uniques": {
                "title": "% Duplicates",
                "description": "Percent Duplicate Di-Tags",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn-rev",
                "modify": lambda x: 100 - x,
            },
            "Valid_Pairs": {
                "title": f"{config.read_count_prefix} Valid",
                "description": f"Valid Pairs ({config.read_count_desc})",
                "min": 0,
                "scale": "PuRd",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
            },
            "Percentage_Valid": {
                "title": "% Valid",
                "description": "Percent Valid Pairs",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
            },
            "Paired_Read_1": {
                "title": f"{config.read_count_prefix} Pairs Aligned",
                "description": f"Paired Alignments ({config.read_count_desc})",
                "min": 0,
                "scale": "PuRd",
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
            },
            "Percentage_Mapped": {
                "title": "% Aligned",
                "description": "Percentage of Paired Alignments",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
            },
        }
        self.general_stats_addcols(self.hicup_data, headers)

    def hicup_truncating_chart(self):
        """Generate the HiCUP Truncated reads plot"""

        # Specify the order of the different possible categories
        keys = {
            "Not_Truncated_Reads": {"color": "#2f7ed8", "name": "Not Truncated"},
            "Truncated_Read": {"color": "#0d233a", "name": "Truncated"},
        }

        # Construct a data structure for the plot - duplicate the samples for read 1 and read 2
        data = {}
        for s_name in self.hicup_data:
            data[f"{s_name} Read 1"] = {}
            data[f"{s_name} Read 2"] = {}
            data[f"{s_name} Read 1"]["Not_Truncated_Reads"] = self.hicup_data[s_name]["Not_Truncated_Reads_1"]
            data[f"{s_name} Read 2"]["Not_Truncated_Reads"] = self.hicup_data[s_name]["Not_Truncated_Reads_2"]
            data[f"{s_name} Read 1"]["Truncated_Read"] = self.hicup_data[s_name]["Truncated_Read_1"]
            data[f"{s_name} Read 2"]["Truncated_Read"] = self.hicup_data[s_name]["Truncated_Read_2"]

        # Config for the plot
        config = {
            "id": "hicup_truncated_reads_plot",
            "title": "HiCUP: Truncated Reads",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        return bargraph.plot(data, keys, config)

    def hicup_alignment_chart(self):
        """Generate the HiCUP Aligned reads plot"""

        # Specify the order of the different possible categories
        keys = {
            "Unique_Alignments_Read": {"color": "#2f7ed8", "name": "Unique Alignments"},
            "Multiple_Alignments_Read": {"color": "#492970", "name": "Multiple Alignments"},
            "Failed_To_Align_Read": {"color": "#0d233a", "name": "Failed To Align"},
            "Too_Short_To_Map_Read": {"color": "#f28f43", "name": "Too short to map"},
        }

        # Construct a data structure for the plot - duplicate the samples for read 1 and read 2
        data = {}
        for s_name in self.hicup_data:
            data[f"{s_name} Read 1"] = {}
            data[f"{s_name} Read 2"] = {}
            data[f"{s_name} Read 1"]["Unique_Alignments_Read"] = self.hicup_data[s_name]["Unique_Alignments_Read_1"]
            data[f"{s_name} Read 2"]["Unique_Alignments_Read"] = self.hicup_data[s_name]["Unique_Alignments_Read_2"]
            data[f"{s_name} Read 1"]["Multiple_Alignments_Read"] = self.hicup_data[s_name]["Multiple_Alignments_Read_1"]
            data[f"{s_name} Read 2"]["Multiple_Alignments_Read"] = self.hicup_data[s_name]["Multiple_Alignments_Read_2"]
            data[f"{s_name} Read 1"]["Failed_To_Align_Read"] = self.hicup_data[s_name]["Failed_To_Align_Read_1"]
            data[f"{s_name} Read 2"]["Failed_To_Align_Read"] = self.hicup_data[s_name]["Failed_To_Align_Read_2"]
            data[f"{s_name} Read 1"]["Too_Short_To_Map_Read"] = self.hicup_data[s_name]["Too_Short_To_Map_Read_1"]
            data[f"{s_name} Read 2"]["Too_Short_To_Map_Read"] = self.hicup_data[s_name]["Too_Short_To_Map_Read_2"]

        # Config for the plot
        config = {
            "id": "hicup_mapping_stats_plot",
            "title": "HiCUP: Mapping Statistics",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        return bargraph.plot(data, keys, config)

    def hicup_filtering_chart(self):
        """Generate the HiCUP filtering plot"""

        # Specify the order of the different possible categories
        keys = {
            "Valid_Pairs": {"color": "#2f7ed8", "name": "Valid Pairs"},
            "Same_Fragment_Internal": {"color": "#0d233a", "name": "Same Fragment - Internal"},
            "Same_Circularised": {"color": "#910000", "name": "Same Fragment - Circularised"},
            "Same_Dangling_Ends": {"color": "#8bbc21", "name": "Same Fragment - Dangling Ends"},
            "Re_Ligation": {"color": "#1aadce", "name": "Re-ligation"},
            "Contiguous_Sequence": {"color": "#f28f43", "name": "Contiguous Sequence"},
            "Wrong_Size": {"color": "#492970", "name": "Wrong Size"},
        }

        # Config for the plot
        config = {
            "id": "hicup_filtering_plot",
            "title": "HiCUP: Filtering Statistics",
            "ylab": "# Read Pairs",
            "cpswitch_counts_label": "Number of Read Pairs",
            "cpswitch_c_active": False,
        }

        return bargraph.plot(self.hicup_data, keys, config)

    def hicup_dedup_chart(self):
        """Generate the HiCUP Deduplication plot"""

        # Specify the order of the different possible categories
        keys = {
            "Deduplication_Cis_Close_Uniques": {"color": "#2f7ed8", "name": "Unique: cis < 10Kbp"},
            "Deduplication_Cis_Far_Uniques": {"color": "#0d233a", "name": "Unique: cis > 10Kbp"},
            "Deduplication_Trans_Uniques": {"color": "#492970", "name": "Unique: trans"},
            "Duplicate_Read_Pairs": {"color": "#f28f43", "name": "Duplicate read pairs"},
        }

        # Config for the plot
        config = {
            "id": "hicup_dedup_plot",
            "title": "HiCUP: De-Duplication Statistics",
            "ylab": "# Di-Tags",
            "cpswitch_counts_label": "Number of Di-Tags",
            "cpswitch_c_active": False,
        }

        return bargraph.plot(self.hicup_data, keys, config)



""" MultiQC module to parse relatedness output from vcftools relatedness """

import csv
import logging
from collections import defaultdict

from multiqc.plots import heatmap

# Initialise the logger
log = logging.getLogger(__name__)


class Relatedness2Mixin:
    def parse_relatedness2(self):
        matrices = {}
        for f in self.find_log_files("vcftools/relatedness2", filehandles=True):
            m = _Relatedness2Matrix(f)
            if m.data and m.x_labels and m.y_labels:
                matrices[f["s_name"]] = m
            self.add_data_source(f, section="Relatedness")

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        matrices = self.ignore_samples(matrices)

        if len(matrices) == 0:
            return 0

        log.info(f"Found {len(matrices)} valid relatedness2 matrices")

        # The matrices cannot be written to a file in their current format
        # self.write_data_file(matrices, "vcftools_relatedness")

        helptext = """
        `RELATEDNESS_PHI` gives a relatedness score between two samples. A higher score indicates a higher degree of
        relatedness, up to a maximum of 0.5. Samples are sorted alphabetically on each axis, and specific IDs can be
        found in the graph with the Highlight tab.
        """

        idx = 0
        for name, m in matrices.items():
            idx += 1
            self.add_section(
                name="Relatedness2",
                anchor=f"vcftools-relatedness2-{idx}",
                description="**Input:** `{}`.\n\n Heatmap of `RELATEDNESS_PHI` values from the output of vcftools relatedness2.".format(
                    name
                ),
                helptext=helptext,
                plot=heatmap.plot(
                    m.data,
                    xcats=m.x_labels,
                    ycats=m.y_labels,
                    pconfig={
                        "id": f"vcftools-relatedness2-heatmap-{idx}",
                        "title": "VCFTools: Relatedness2",
                        "square": True,
                        "decimalPlaces": 7,
                    },
                ),
            )

        return len(matrices)


class _Relatedness2Matrix:
    def __init__(self, relatedness_file):
        self.data = []
        self.x_labels = set()
        self.y_labels = set()

        self.parse(relatedness_file["f"])

    def parse(self, f):
        rels = defaultdict(dict)
        r = csv.DictReader(f, delimiter="\t")
        for line in r:
            self.x_labels.add(line["INDV1"])
            self.y_labels.add(line["INDV2"])

            rels[line["INDV1"]][line["INDV2"]] = float(line["RELATEDNESS_PHI"])

        # impose alphabetical order and avoid json serialisation errors in utils.report
        self.x_labels = sorted(self.x_labels)
        self.y_labels = sorted(self.y_labels)

        for x in self.x_labels:
            line = []
            for y in self.y_labels:
                line.append(rels[x][y])
            self.data.append(line)



""" MultiQC module to parse output from Trimmomatic """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Trimmomatic",
            anchor="trimmomatic",
            href="http://www.usadellab.org/cms/?page=trimmomatic",
            info="is a flexible read trimming tool for Illumina NGS data.",
            doi="10.1093/bioinformatics/btu170",
        )

        # Parse logs
        self.trimmomatic = dict()
        for f in self.find_log_files("trimmomatic", filehandles=True):
            self.parse_trimmomatic(f)
            self.add_data_source(f)

        # Filter to strip out ignored sample names
        self.trimmomatic = self.ignore_samples(self.trimmomatic)

        if len(self.trimmomatic) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(self.trimmomatic)} logs")

        self.write_data_file(self.trimmomatic, "multiqc_trimmomatic")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Add drop rate to the general stats table
        headers = {
            "dropped_pct": {
                "title": "% Dropped",
                "description": "% Dropped reads",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "OrRd",
            }
        }
        self.general_stats_addcols(self.trimmomatic, headers)

        # Make barplot
        self.trimmomatic_barplot()

    def parse_trimmomatic(self, f):
        s_name = None
        if getattr(config, "trimmomatic", {}).get("s_name_filenames", False):
            s_name = f["s_name"]
        for line in f["f"]:
            # Get the sample name
            if s_name is None and line.startswith(
                tuple(f"Trimmomatic{x}E: Started with arguments:" for x in ["S", "P"])
            ):
                is_pe = line.startswith("TrimmomaticPE")
                args = line.strip().split()
                FQ_EXTS = ".fastq", ".fq", ".gz", ".dat"
                if not any(x.endswith(FQ_EXTS) for x in args):
                    # Try looking on the next line instead, sometimes have a line break (see issue #212)
                    line = next(f["f"])
                    args = line.strip().split()
                if any(x.endswith(FQ_EXTS) for x in args):
                    # For PE, first two fastq files are the input paths; for SE, it's just the first one
                    input_paths = [x for x in args if x.endswith(FQ_EXTS)][: 2 if is_pe else 1]
                    s_name = self.clean_s_name(input_paths, f)
                    if s_name in self.trimmomatic:
                        log.debug(f"Duplicate sample name found! Overwriting: {s_name}")

            # Get single end stats
            if "Input Reads" in line and s_name is not None:
                match = re.search(
                    r"Input Reads: (\d+) Surviving: (\d+) \(([\d\.,]+)%\) Dropped: (\d+) \(([\d\.,]+)%\)", line
                )
                if match:
                    self.trimmomatic[s_name] = {
                        "input_reads": float(match.group(1)),
                        "surviving": float(match.group(2)),
                        "surviving_pct": float(match.group(3).replace(",", ".")),
                        "dropped": float(match.group(4)),
                        "dropped_pct": float(match.group(5).replace(",", ".")),
                    }
                    s_name = None

            # Get paired end stats
            if "Input Read Pairs" in line and s_name is not None:
                match = re.search(
                    r"Input Read Pairs: (\d+) Both Surviving: (\d+) \(([\d\.,]+)%\) Forward Only Surviving: (\d+) \(([\d\.,]+)%\) Reverse Only Surviving: (\d+) \(([\d\.,]+)%\) Dropped: (\d+) \(([\d\.,]+)%\)",
                    line,
                )
                if match:
                    self.trimmomatic[s_name] = {
                        "input_read_pairs": float(match.group(1)),
                        "surviving": float(match.group(2)),
                        "surviving_pct": float(match.group(3).replace(",", ".")),
                        "forward_only_surviving": float(match.group(4)),
                        "forward_only_surviving_pct": float(match.group(5).replace(",", ".")),
                        "reverse_only_surviving": float(match.group(6)),
                        "reverse_only_surviving_pct": float(match.group(7).replace(",", ".")),
                        "dropped": float(match.group(8)),
                        "dropped_pct": float(match.group(9).replace(",", ".")),
                    }
                    s_name = None

    def trimmomatic_barplot(self):
        """Make the HighCharts HTML to plot the trimmomatic rates"""

        # Specify the order of the different possible categories
        keys = {
            "surviving": {"color": "#437bb1", "name": "Surviving Reads"},
            "both_surviving": {"color": "#f7a35c", "name": "Both Surviving"},
            "forward_only_surviving": {"color": "#e63491", "name": "Forward Only Surviving"},
            "reverse_only_surviving": {"color": "#b1084c", "name": "Reverse Only Surviving"},
            "dropped": {"color": "#7f0000", "name": "Dropped"},
        }

        # Config for the plot
        pconfig = {
            "id": "trimmomatic_plot",
            "title": "Trimmomatic: Surviving Reads",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        self.add_section(plot=bargraph.plot(self.trimmomatic, keys, pconfig))



""" MultiQC submodule to parse output from RSeQC infer_experiment.py
http://rseqc.sourceforge.net/#infer-experiment-py """

import logging
import re

from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find RSeQC infer_experiment reports and parse their data"""

    # Set up vars
    self.infer_exp = dict()
    regexes = {
        "pe_sense": r"\"1\+\+,1--,2\+-,2-\+\": (\d\.\d+)",
        "pe_antisense": r"\"1\+-,1-\+,2\+\+,2--\": (\d\.\d+)",
        "se_sense": r"\"\+\+,--\": (\d\.\d+)",
        "se_antisense": r"\+-,-\+\": (\d\.\d+)",
        "failed": r"Fraction of reads failed to determine: (\d\.\d+)",
    }

    # Go through files and parse data using regexes
    for f in self.find_log_files("rseqc/infer_experiment"):
        d = dict()
        for k, r in regexes.items():
            r_search = re.search(r, f["f"], re.MULTILINE)
            if r_search:
                d[k] = float(r_search.group(1))

        if len(d) > 0:
            if f["s_name"] in self.infer_exp:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.add_data_source(f, section="infer_experiment")
            self.infer_exp[f["s_name"]] = d

    # Filter to strip out ignored sample names
    self.infer_exp = self.ignore_samples(self.infer_exp)

    if len(self.infer_exp) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Write to file
    self.write_data_file(self.infer_exp, "multiqc_rseqc_infer_experiment")

    # Merge PE and SE for plot
    pdata = dict()
    for s_name, vals in self.infer_exp.items():
        pdata[s_name] = dict()
        for k, v in vals.items():
            v *= 100.0  # Multiply to get percentage
            if k[:2] == "pe" or k[:2] == "se":
                k = k[3:]
            pdata[s_name][k] = v + pdata[s_name].get(k, 0)

    # Plot bar graph of groups
    keys = {
        "sense": {"name": "Sense"},
        "antisense": {"name": "Antisense"},
        "failed": {"name": "Undetermined"},
    }
    # Config for the plot
    pconfig = {
        "id": "rseqc_infer_experiment_plot",
        "title": "RSeQC: Infer experiment",
        "ylab": "% Tags",
        "ymin": 0,
        "ymax": 100,
        "tt_percentages": False,
        "ylab_format": "{value}%",
        "cpswitch": False,
    }

    self.add_section(
        name="Infer experiment",
        anchor="rseqc-infer_experiment",
        description='<a href="http://rseqc.sourceforge.net/#infer-experiment-py" target="_blank">Infer experiment</a>'
        " counts the percentage of reads and read pairs that match the strandedness of overlapping transcripts."
        " It can be used to infer whether RNA-seq library preps are stranded (sense or antisense).",
        plot=bargraph.plot(pdata, keys, pconfig),
    )

    # Return number of samples found
    return len(self.infer_exp)



#!/usr/bin/env python

""" MultiQC datatable class, used by tables and beeswarm plots """

import logging
import random
import re
import string
from collections import defaultdict
from typing import List, Tuple, Dict

from multiqc.utils import config, report

logger = logging.getLogger(__name__)


class DataTable:
    """Data table class. Prepares and holds data and configuration
    for either a table or a beeswarm plot."""

    def __init__(self, data, headers=None, pconfig=None):
        """Prepare data for use in a table or plot"""
        if headers is None:
            headers = []
        if pconfig is None:
            pconfig = {}

        # Allow user to overwrite any given config for this plot
        if pconfig.get("id") and pconfig["id"] in config.custom_plot_config:
            for k, v in config.custom_plot_config[pconfig["id"]].items():
                pconfig[k] = v

        if not pconfig.get("id"):
            if config.strict:
                errmsg = f"LINT: 'id' is missing from plot pconfig: {pconfig}"
                logger.error(errmsg)
                report.lint_errors.append(errmsg)
            pconfig["id"] = report.save_htmlid(f"table_{''.join(random.sample(string.ascii_lowercase, 4))}")

        # Given one dataset - turn it into a list
        if not isinstance(data, list):
            data = [data]
        if not isinstance(headers, list):
            headers = [headers]

        sectcols = [
            "55,126,184",
            "77,175,74",
            "152,78,163",
            "255,127,0",
            "228,26,28",
            "179,179,50",
            "166,86,40",
            "247,129,191",
            "153,153,153",
        ]

        # Go through each table section
        for idx, d in enumerate(data):
            # Get the header keys
            try:
                keys = headers[idx].keys()
                assert len(keys) > 0
            except (IndexError, AttributeError, AssertionError):
                pconfig["only_defined_headers"] = False

            # Add header keys from the data
            if pconfig.get("only_defined_headers", True) is False:
                # Get the keys from the data
                keys = list()
                for samp in d.values():
                    for k in samp.keys():
                        if k not in keys:
                            keys.append(k)

                # If we don't have a headers dict for this data set yet, create one
                try:
                    headers[idx]
                except IndexError:
                    headers.append(list)
                    headers[idx] = dict()
                else:
                    # Convert the existing headers into a dict (e.g. if parsed from a config)
                    od_tuples = [(key, headers[idx][key]) for key in headers[idx].keys()]
                    headers[idx] = dict(od_tuples)

                # Create empty header configs for each new data key
                for k in keys:
                    if k not in headers[idx]:
                        headers[idx][k] = {}

            # Ensure that keys are strings, not numeric
            keys = [str(k) for k in keys]
            for k in list(headers[idx].keys()):
                headers[idx][str(k)] = headers[idx].pop(k)
            # Ensure that all sample names are strings as well
            cdata = dict()
            for k, v in data[idx].items():
                cdata[str(k)] = v
            data[idx] = cdata
            for s_name in data[idx].keys():
                for k in list(data[idx][s_name].keys()):
                    data[idx][s_name][str(k)] = data[idx][s_name].pop(k)

            # Check that we have some data in each column
            empties = list()
            for k in keys:
                n = 0
                for samp in d.values():
                    if k in samp:
                        n += 1
                if n == 0:
                    empties.append(k)
            for k in empties:
                keys = [j for j in keys if j != k]
                del headers[idx][k]

            for k in keys:
                # Unique id to avoid overwriting by other datasets
                if "rid" not in headers[idx][k]:
                    headers[idx][k]["rid"] = report.save_htmlid(re.sub(r"\W+", "_", k).strip().strip("_"))

                # Applying defaults presets for data keys if shared_key is set to base_count or read_count
                shared_key = headers[idx][k].get("shared_key", None)
                suffix = None
                if shared_key in ["read_count", "long_read_count", "base_count"]:
                    if shared_key == "read_count":
                        multiplier = config.read_count_multiplier
                        suffix = " " + config.read_count_prefix
                    elif shared_key == "long_read_count":
                        multiplier = config.long_read_count_multiplier
                        suffix = " " + config.long_read_count_prefix
                    elif shared_key == "base_count":
                        multiplier = config.base_count_multiplier
                        suffix = " " + config.base_count_prefix
                    if headers[idx][k].get("modify") is None:
                        headers[idx][k]["modify"] = lambda x: x * multiplier
                    if headers[idx][k].get("min") is None:
                        headers[idx][k]["min"] = 0
                    if headers[idx][k].get("format") is None:
                        if multiplier == 1:
                            headers[idx][k]["format"] = "{:,.0f}"
                if suffix:
                    suffix = suffix.replace(" ", "&nbsp;")
                    if "suffix" not in headers[idx][k]:
                        headers[idx][k]["suffix"] = suffix

                # Use defaults / data keys if headers not given
                headers[idx][k]["namespace"] = headers[idx][k].get("namespace", pconfig.get("namespace", ""))
                headers[idx][k]["title"] = headers[idx][k].get("title", k)
                headers[idx][k]["description"] = headers[idx][k].get("description", headers[idx][k]["title"])
                headers[idx][k]["scale"] = headers[idx][k].get("scale", pconfig.get("scale", "GnBu"))
                headers[idx][k]["format"] = headers[idx][k].get("format", pconfig.get("format", "{:,.1f}"))
                headers[idx][k]["colour"] = headers[idx][k].get("colour", pconfig.get("colour", None))
                headers[idx][k]["hidden"] = headers[idx][k].get("hidden", pconfig.get("hidden", None))
                headers[idx][k]["max"] = headers[idx][k].get("max", pconfig.get("max", None))
                headers[idx][k]["min"] = headers[idx][k].get("min", pconfig.get("min", None))
                headers[idx][k]["ceiling"] = headers[idx][k].get("ceiling", pconfig.get("ceiling", None))
                headers[idx][k]["floor"] = headers[idx][k].get("floor", pconfig.get("floor", None))
                headers[idx][k]["minRange"] = headers[idx][k].get("minRange", pconfig.get("minRange", None))
                headers[idx][k]["shared_key"] = headers[idx][k].get("shared_key", pconfig.get("shared_key", None))
                headers[idx][k]["modify"] = headers[idx][k].get("modify", pconfig.get("modify", None))
                headers[idx][k]["placement"] = float(headers[idx][k].get("placement", 1000))

                if headers[idx][k]["colour"] is None:
                    cidx = idx
                    while cidx >= len(sectcols):
                        cidx -= len(sectcols)
                    headers[idx][k]["colour"] = sectcols[cidx]

                # Overwrite (2nd time) any given config with table-level user config
                # This is to override column-specific values set by modules
                if "id" in pconfig and pconfig["id"] and pconfig["id"] in config.custom_plot_config:
                    for cpc_k, cpc_v in config.custom_plot_config[pconfig["id"]].items():
                        headers[idx][k][cpc_k] = cpc_v

                # Overwrite "name" if set in user config
                # Key can be a column ID, a table ID, or a namespace in the general stats table.
                for key, val in config.table_columns_name.items():
                    key = key.lower()
                    # Case-insensitive check if the outer key is a table ID or a namespace.
                    if key in [pconfig["id"].lower(), headers[idx][k]["namespace"].lower()] and isinstance(val, dict):
                        # Assume a dict of specific column IDs
                        for key2, new_title in val.items():
                            key2 = key2.lower()
                            if key2 in [k.lower(), headers[idx][k]["title"].lower()]:
                                headers[idx][k]["title"] = new_title

                    # Case-insensitive check if the outer key is a column ID
                    elif key in [k.lower(), headers[idx][k]["title"].lower()] and isinstance(val, str):
                        headers[idx][k]["title"] = val

                # Overwrite "hidden" if set in user config
                # Key can be a column ID, a table ID, or a namespace in the general stats table.
                for key, val in config.table_columns_visible.items():
                    key = key.lower()
                    # Case-insensitive check if the outer key is a table ID or a namespace.
                    if key in [pconfig["id"].lower(), headers[idx][k]["namespace"].lower()]:
                        # First - if config value is a bool, set all module columns to that value
                        if isinstance(val, bool):
                            # Config has True = visible, False = Hidden. Here we're setting "hidden" which is inverse
                            headers[idx][k]["hidden"] = not val

                        # Not a bool, assume a dict of specific column IDs
                        elif isinstance(val, dict):
                            for key2, visible in val.items():
                                key2 = key2.lower()
                                if key2 in [k.lower(), headers[idx][k]["title"].lower()] and isinstance(visible, bool):
                                    # Config has True = visible, False = Hidden. Here we're setting "hidden" which is inverse
                                    headers[idx][k]["hidden"] = not visible

                    # Case-insensitive check if the outer key is a column ID
                    elif key in [k.lower(), headers[idx][k]["title"].lower()] and isinstance(val, bool):
                        # Config has True = visible, False = Hidden. Here we're setting "hidden" which is inverse
                        headers[idx][k]["hidden"] = not val

                # Also overwrite placement if set in config
                try:
                    headers[idx][k]["placement"] = float(
                        config.table_columns_placement[headers[idx][k]["namespace"]][k]
                    )
                except (KeyError, ValueError):
                    try:
                        headers[idx][k]["placement"] = float(config.table_columns_placement[pconfig["id"]][k])
                    except (KeyError, ValueError):
                        pass

                # Overwrite any header config if set in config
                for custom_k, custom_v in (
                    config.custom_table_header_config.get(pconfig.get("id"), {}).get(k, {}).items()
                ):
                    headers[idx][k][custom_k] = custom_v

                # Work out max and min value if not given
                setdmax = False
                setdmin = False
                try:
                    headers[idx][k]["dmax"] = float(headers[idx][k]["max"])
                except Exception:
                    headers[idx][k]["dmax"] = 0
                    setdmax = True

                try:
                    headers[idx][k]["dmin"] = float(headers[idx][k]["min"])
                except Exception:
                    headers[idx][k]["dmin"] = 0
                    setdmin = True

                # Figure out the min / max if not supplied
                if setdmax or setdmin:
                    for s_name, samp in data[idx].items():
                        try:
                            val = float(samp[k])
                            if callable(headers[idx][k]["modify"]):
                                val = float(headers[idx][k]["modify"](val))
                            if setdmax:
                                headers[idx][k]["dmax"] = max(headers[idx][k]["dmax"], val)
                            if setdmin:
                                headers[idx][k]["dmin"] = min(headers[idx][k]["dmin"], val)
                        except (ValueError, TypeError):
                            val = samp[k]  # couldn't convert to float - keep as a string
                        except KeyError:
                            pass  # missing data - skip
                    # Limit auto-generated scales with floor, ceiling and minRange.
                    if headers[idx][k]["ceiling"] is not None and headers[idx][k]["max"] is None:
                        headers[idx][k]["dmax"] = min(headers[idx][k]["dmax"], float(headers[idx][k]["ceiling"]))
                    if headers[idx][k]["floor"] is not None and headers[idx][k]["min"] is None:
                        headers[idx][k]["dmin"] = max(headers[idx][k]["dmin"], float(headers[idx][k]["floor"]))
                    if headers[idx][k]["minRange"] is not None:
                        drange = headers[idx][k]["dmax"] - headers[idx][k]["dmin"]
                        if drange < float(headers[idx][k]["minRange"]):
                            headers[idx][k]["dmax"] = headers[idx][k]["dmin"] + float(headers[idx][k]["minRange"])

        # Collect settings for shared keys
        shared_keys = defaultdict(lambda: dict())
        for idx, hs in enumerate(headers):
            for k in hs.keys():
                sk = headers[idx][k]["shared_key"]
                if sk is not None:
                    shared_keys[sk]["dmax"] = max(
                        headers[idx][k]["dmax"], shared_keys[sk].get("dmax", headers[idx][k]["dmax"])
                    )
                    shared_keys[sk]["dmin"] = min(
                        headers[idx][k]["dmin"], shared_keys[sk].get("dmin", headers[idx][k]["dmin"])
                    )

        # Overwrite shared key settings and at the same time assign to buckets for sorting
        # So the final ordering is:
        #   placement > section > explicit_ordering
        # Of course, the user can shuffle these manually.
        self.headers_in_order = defaultdict(list)
        for idx, hs in enumerate(headers):
            for k in hs.keys():
                sk = headers[idx][k]["shared_key"]
                if sk is not None:
                    headers[idx][k]["dmax"] = shared_keys[sk]["dmax"]
                    headers[idx][k]["dmin"] = shared_keys[sk]["dmin"]

                self.headers_in_order[headers[idx][k]["placement"]].append((idx, k))

        # Skip any data that is not used in the table
        # Would be ignored for making the table anyway, but can affect whether a beeswarm plot is used
        for idx, d in enumerate(data):
            for s_name in list(d.keys()):
                if not any(h in data[idx][s_name].keys() for h in headers[idx]):
                    del data[idx][s_name]

        # Assign to class
        self.data = data
        self.headers = headers
        self.pconfig = pconfig

    def get_headers_in_order(self) -> List[Tuple[int, str, Dict]]:
        """
        Gets the headers in the order they want to be displayed.
        Returns a list of triplets: (bucket_idx, key, header_info)
        """
        res = list()
        # Scan through self.headers_in_order and just bolt on the actual header info
        for bucket in sorted(self.headers_in_order):
            for bucket_idx, k in self.headers_in_order[bucket]:
                res.append((bucket_idx, k, self.headers[bucket_idx][k]))
        return res



import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="SeqyClean",
            anchor="seqyclean",
            href="https://github.com/ibest/seqyclean",
            info="is a pre-processing tool for NGS data that filters adapters, vectors, and contaminants while quality trimming.",
            doi="10.1145/3107411.3107446",
        )

        # Parse logs
        self.seqyclean_data = dict()
        for f in self.find_log_files("seqyclean"):
            rows = f["f"].splitlines()
            headers = rows[0].split("\t")
            cols = rows[1].split("\t")

            self.seqyclean_data[f["s_name"]] = dict()
            for header, col in zip(headers, cols):
                # Add verions info
                if header == "Version":
                    self.add_software_version(col, f["s_name"])

                # Attempt to convert into a float if we can
                try:
                    col = float(col)
                except (ValueError, TypeError):
                    pass
                self.seqyclean_data[f["s_name"]].update({header: col})

            self.add_data_source(f)

        if len(self.seqyclean_data) == 0:
            raise ModuleNoSamplesFound

        self.seqyclean_data = self.ignore_samples(self.seqyclean_data)

        if len(self.seqyclean_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.seqyclean_data)} logs")

        # Adding the bar plots
        self.add_section(
            name="Summary",
            anchor="seqyclean-summary",
            description="This plot shows the number of reads that were kept and discarded.",
            plot=self.seqyclean_summary(),
        )

        self.add_section(
            name="Annotations",
            anchor="seqyclean-annotation",
            description="This plot shows how reads were annotated.",
            plot=self.seqyclean_analysis(),
        )

        self.add_section(
            name="Discarded",
            anchor="seqyclean-discarded",
            description="This plot shows the breakdown of reasons for why reads were discarded.",
            plot=self.seqyclean_discarded(),
        )

        # Write the results to a file
        self.write_data_file(self.seqyclean_data, "multiqc_seqyclean")

        # Adding to the general statistics table
        self.seqyclean_general_stats_table()

    def seqyclean_summary(self):
        config = {"id": "seqyclean-summary-plot", "title": "SeqyClean: Summary", "ylab": "Number of Reads"}
        keys = [
            "PairsKept",  # for paired end
            "PairsDiscarded",
            "PE1DiscardedTotal",
            "PE2DiscardedTotal",
            "SEReadsKept",  # single end
            "SEDiscardedTotal",
            "ReadsKept",  # 454
            "DiscardedTotal",
        ]
        return bargraph.plot(self.seqyclean_data, self._clean_keys(keys), config)

    def seqyclean_analysis(self):
        config = {
            "id": "seqyclean-read-annotation-plot",
            "title": "SeqyClean: Read Annotations",
            "ylab": "Number of Reads",
        }
        keys = [
            "PE1TruSeqAdap_found",  # for paired end
            "PE1ReadsWVector_found",
            "PE1ReadsWContam_found",
            "PE2TruSeqAdap_found",
            "PE2ReadsWVector_found",
            "PE2ReadsWContam_found",
            "SETruSeqAdap_found",  # for single end
            "SEReadsWVector_found",
            "SEReadsWContam_found",
            "left_mid_tags_found",  # 454 data
            "right_mid_tags_found",
            "ReadsWithVector_found",
            "ReadsWithContam_found",
        ]
        return bargraph.plot(self.seqyclean_data, self._clean_keys(keys), config)

    def seqyclean_discarded(self):
        config = {
            "id": "seqyclean-discarded-reads-plot",
            "title": "SeqyClean: Discarded Reads",
            "ylab": "Number of Reads",
        }
        keys = [
            "SEDiscByContam",  # single end
            "SEDiscByLength",
            "PE1DiscByContam",  # paired end
            "PE1DiscByLength",
            "PE2DiscByContam",
            "PE2DiscByLength",
            "DiscByContam",  # 454 data
            "DiscByLength",
        ]
        return bargraph.plot(self.seqyclean_data, self._clean_keys(keys), config)

    def seqyclean_general_stats_table(self):
        headers = {
            "Perc_Kept": {
                "title": "% Kept",
                "description": "The percentage of reads remaining after cleaning",
                "scale": "YlGn",
                "suffix": "%",
                "max": 100,
                "min": 0,
            },
            "PercentageKept": {
                "title": "% Kept",
                "description": "The percentage of reads remaining after cleaning",
                "scale": "YlGn",
                "suffix": "%",
                "max": 100,
                "min": 0,
            },
        }
        self.general_stats_addcols(self.seqyclean_data, headers)

    def _clean_keys(self, keys):
        """Given a list of keys, make them easier to read for plot labels"""
        cats = {}
        for k in keys:
            nice_name = re.sub(r"([a-z])([A-Z])", r"\g<1> \g<2>", k)  # CamelCase > Camel Case
            nice_name = re.sub(r"([PS]E\d?)", r"\g<1> ", nice_name)  # PE1Label > PE1 Label
            nice_name = re.sub(r"W([A-Z])", r"W \g<1>", nice_name)  # WContam > W Contam
            nice_name = nice_name.replace("_", " ")  # tags_found > tags found
            nice_name = nice_name.title()  # Title Case
            nice_name = nice_name.replace("Pe", "PE").replace("Se", "SE")
            nice_name = nice_name.replace("Tru SEq", "TruSeq")
            cats[k] = {"name": nice_name}
        return cats



""" MultiQC submodule to parse output from Picard QualityScoreDistribution """

import logging
from collections import OrderedDict

from multiqc.plots import linegraph

from .util import read_histogram

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find Picard QualityScoreDistribution reports and parse their data"""

    headers = ["QUALITY", "COUNT_OF_Q"]
    formats = [int, int]
    all_data = read_histogram(
        self,
        "picard/quality_score_distribution",
        headers,
        formats,
        picard_tool="QualityScoreDistribution",
        sentieon_algo="QualDistribution",
    )

    if not all_data:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Write parsed data to a file
    self.write_data_file(all_data, f"multiqc_{self.anchor}_quality_score_distribution")

    # Plot the data and add section
    pconfig = {
        "id": f"{self.anchor}_quality_score_distribution",
        "title": f"{self.name}: Base Quality Distribution",
        "ylab": "Number of Bases",
        "xlab": "Base Quality Score",
        "xDecimals": False,
        "tt_label": "<b>base quality{point.x}</b>: {point.y}",
        "ymin": 0,
    }

    lg = {}
    for s_name in all_data:
        lg[s_name] = OrderedDict((qual, data["COUNT_OF_Q"]) for qual, data in all_data[s_name].items())

    self.add_section(
        name="Base Quality Distribution",
        anchor=f"{self.anchor}-quality-score-distribution",
        description="Plot shows the count of each base quality score.",
        plot=linegraph.plot([lg], pconfig),
    )

    # Return the number of detected samples to the parent module
    return len(all_data)



from itertools import chain

from multiqc.plots import linegraph


def plot_bhist(samples, file_type, **plot_args):
    """Create line graph plot of histogram data for BBMap 'bhist' output.

    The 'samples' parameter could be from the bbmap mod_data dictionary:
    samples = bbmap.MultiqcModule.mod_data[file_type]
    """

    all_x = set()
    for item in sorted(chain(*[samples[sample]["data"].items() for sample in samples])):
        all_x.add(item[0])

    columns_to_plot = {
        "GC": {
            1: "C",
            2: "G",
        },
        "AT": {
            0: "A",
            3: "T",
        },
        "N": {4: "N"},
    }
    nucleotide_data = []
    for column_type in columns_to_plot:
        nucleotide_data.append(
            {
                sample + "." + column_name: {
                    x: samples[sample]["data"][x][column] * 100 if x in samples[sample]["data"] else 0 for x in all_x
                }
                for sample in samples
                for column, column_name in columns_to_plot[column_type].items()
            }
        )

    plot_params = {
        "id": "bbmap-" + file_type + "_plot",
        "title": "BBTools: " + plot_args["plot_title"],
        "xlab": "Read position",
        "ylab": "Percentage of G+C bases",
        "ymin": 0,
        "ymax": 100,
        "data_labels": [
            {"name": "Percentage of G+C bases"},
            {"name": "Percentage of A+T bases"},
            {"name": "Percentage of N bases"},
        ],
    }
    plot_params.update(plot_args["plot_params"])
    plot = linegraph.plot(nucleotide_data, plot_params)

    return plot



import logging
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph, table

log = logging.getLogger(__name__)


class DragenGcMetrics(BaseMultiqcModule):
    """Not to be confused with DragenFastqcGcMetrics"""

    NAMESPACE = "GC Metrics"

    def add_gc_metrics_hist(self):
        data_by_sample = dict()

        for f in self.find_log_files("dragen/gc_metrics"):
            data = parse_gc_metrics_file(f)
            s_name = f["s_name"]
            if s_name in data_by_sample:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, section="gc_metrics")
            data_by_sample[s_name] = data

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, s_name)

        # Filter to strip out ignored sample names:
        data_by_sample = self.ignore_samples(data_by_sample)

        if not data_by_sample:
            return set()

        # Only plot data, don't want to write this to a file
        # (can do so with --export-plots already)
        # self.write_data_file(data_by_sample, "dragen_gc_metrics")

        hist_data = DragenGcMetrics.__get_normalized_gc_data(data_by_sample)
        smooth_points = 300
        self.add_section(
            name="GC Bias Histogram",
            anchor="dragen-gc-bias-hist",
            description="""
                A histogram of the normalized coverage vs GC content.  This shows how GC
                content in the genome impacts sequencing coverage.
                """,
            plot=linegraph.plot(
                hist_data,
                {
                    "id": "gc-bias-hist",
                    "title": "Dragen: GC Bias Histogram",
                    "ylab": "Normalized Coverage",
                    "xlab": "% GC",
                    "ymin": 0,
                    "xmin": 0,
                    "tt_label": "<b>{point.x} % GC</b>: {point.y} Normalized coverage",
                    "smooth_points": smooth_points,
                    "namespace": DragenGcMetrics.NAMESPACE,
                },
            ),
        )

        table_data = DragenGcMetrics.__get_summary_gc_data(data_by_sample)
        self.add_section(
            name="GC Metrics Summary",
            anchor="dragen-gc-metrics-summary",
            description="""
            Summary GC metrics shown on the sample level.
            """,
            plot=table.plot(
                table_data,
                pconfig={
                    "id": "dragen-gc-metrics-summary-table",
                    "namespace": DragenGcMetrics.NAMESPACE,
                },
            ),
        )

        return data_by_sample.keys()

    @staticmethod
    def __get_normalized_gc_data(data_by_sample) -> dict:
        """Returns headers, data"""
        analysis = "GC BIAS DETAILS"
        hist_data = {}
        for sample_name, sample_data in data_by_sample.items():
            # {Normalized coverage at GC 0: 0.8324, Normalized coverage at GC 1,0.9456}
            hist_data[sample_name] = {
                int(key.split()[-1]): item
                for key, item in sample_data[analysis].items()
                if key.startswith("Normalized coverage at GC")
            }
        return hist_data

    @staticmethod
    def __get_summary_gc_data(data_by_sample) -> dict:
        analysis = "GC METRICS SUMMARY"
        summary_data = {}
        for sample_name, sample_data in data_by_sample.items():
            summary_data[sample_name] = {metric: stat for metric, stat in sample_data[analysis].items()}

        return summary_data


def parse_gc_metrics_file(f):
    """
    sample.gc_metrics.csv

    GC BIAS DETAILS,,Windows at GC 0,20,0.001
    GC BIAS DETAILS,,Windows at GC 1,4,0.000
    ...
    GC BIAS DETAILS,,Normalized coverage at GC 0,0.8324
    GC BIAS DETAILS,,Normalized coverage at GC 1,0.8540
    GC BIAS DETAILS,,Normalized coverage at GC 2,0.8512
    ...
    GC METRICS SUMMARY,,Window size,100
    GC METRICS SUMMARY,,Number of valid windows,3118404
    GC METRICS SUMMARY,,Number of discarded windows,366533633
    GC METRICS SUMMARY,,Average reference GC,40.84
    GC METRICS SUMMARY,,Mean global coverage,30.15
    GC METRICS SUMMARY,,Normalized coverage at GCs 0-19,1.06
    GC METRICS SUMMARY,,Normalized coverage at GCs 20-39,1.05
    GC METRICS SUMMARY,,Normalized coverage at GCs 40-59,0.97
    GC METRICS SUMMARY,,Normalized coverage at GCs 60-79,0.84
    GC METRICS SUMMARY,,Normalized coverage at GCs 80-100,0.47
    GC METRICS SUMMARY,,AT Dropout,0.58
    GC METRICS SUMMARY,,GC Dropout,2.01
    """

    data = defaultdict(dict)
    for line in f["f"].splitlines():
        tokens = line.split(",")
        # Percentage is currently unused
        if len(tokens) == 4:
            analysis, _, metric, stat = tokens
            percentage = None
        elif len(tokens) == 5:
            analysis, _, metric, stat, percentage = tokens
        else:
            raise ValueError(f"Unexpected number of tokens in line {line}")

        try:
            stat = float(stat)
        except ValueError:
            pass
        data[analysis][metric] = stat

    return data



""" MultiQC submodule to parse output from Picard OxoGMetrics """

import logging
from collections import defaultdict

from multiqc.modules.picard import util

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(module):
    """Find Picard OxoGMetrics reports and parse their data"""

    # Set up vars
    data_by_sample = dict()

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/oxogmetrics", filehandles=True):
        # Sample name from input file name by default.
        s_name = f["s_name"]
        keys = None
        context_col = None

        for line in f["f"]:
            maybe_s_name = util.extract_sample_name(
                module,
                line,
                f,
                picard_tool=["CollectOxoGMetrics", "ConvertSequencingArtifactToOxoG"],
                picard_opt=["INPUT", "INPUT_BASE"],
            )
            if maybe_s_name:
                s_name = maybe_s_name
                keys = None
                context_col = None

            if s_name is None:
                continue

            if util.is_line_right_before_table(line, picard_class="CollectOxoGMetrics"):
                keys = f["f"].readline().strip("\n").split("\t")
                context_col = keys.index("CONTEXT")
                if s_name in data_by_sample:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: " f"{s_name}")
                data_by_sample[s_name] = defaultdict()
                module.add_data_source(f, s_name, section="OxoGMetrics")

            elif keys:
                vals = line.strip("\n").split("\t")
                if len(vals) != len(keys) or context_col is None:
                    s_name = None
                    keys = None
                    context_col = None
                    continue

                context = vals[context_col]
                data_by_sample[s_name][context] = dict()
                for i, k in enumerate(keys):
                    k = k.strip()
                    try:
                        val = float(vals[i])
                    except ValueError:
                        val = vals[i].strip()
                    data_by_sample[s_name][context][k] = val

    data_by_sample = module.ignore_samples(data_by_sample)
    if len(data_by_sample) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    # Collapse into 2D structure with sample_context keys
    print_data = {f"{s}_{c}": v for s in data_by_sample.keys() for c, v in data_by_sample[s].items()}
    module.write_data_file(print_data, "multiqc_picard_OxoGMetrics")

    # Add to general stats table
    general_stats_data = dict()
    for s_name in data_by_sample:
        general_stats_data[s_name] = dict()
        try:
            val = data_by_sample[s_name]["CCG"]["OXIDATION_ERROR_RATE"]
            general_stats_data[s_name]["CCG_OXIDATION_ERROR_RATE"] = val
        except KeyError:
            log.warning(f"Couldn't find picard CCG oxidation error rate for {s_name}")
    headers = {
        "CCG_OXIDATION_ERROR_RATE": {
            "title": "CCG Oxidation",
            "description": "CCG-CAG Oxidation Error Rate",
            "min": 0,
            "max": 100,
            "suffix": "%",
            "format": "{:,.0f}",
            "scale": "RdYlGn-rev",
            "modify": lambda x: util.multiply_hundred(x),
        }
    }
    module.general_stats_addcols(general_stats_data, headers, namespace="OxoGMetrics")

    # Return the number of detected samples to the parent module
    return len(data_by_sample)



""" MultiQC module to parse output from Anglerfish """

import json
import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, beeswarm, table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    Anglerfish module class
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Anglerfish",
            anchor="Anglerfish",
            href="https://github.com/remiolsen/anglerfish",
            info="A tool to assess Illumina libraries sequenced on Oxford Nanopore for the purpose of quality control.",
            # doi="", No DOI available
        )

        # Find and load any anglerfish reports
        self.anglerfish_data = dict()

        for f in self.find_log_files("anglerfish", filehandles=True):
            self.parse_anglerfish_json(f)

        # Filter to strip out ignored sample names
        self.anglerfish_data = self.ignore_samples(self.anglerfish_data)

        # Stop execution of the data if no anglerfish data is found.
        if len(self.anglerfish_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.anglerfish_data)} reports")

        # Write parsed report data to a file
        ## Parse whole JSON to save all its content
        self.write_data_file(self.anglerfish_data, "multiqc_anglerfish")

        # General Stats Table
        self.anglerfish_general_stats_table()

        # Adds section for Sample Stats Read length table/beeswarm plot
        self.anglerfish_sample_stats()
        # Adds section for Undetermined indexes plot
        self.anglerfish_undetermined_index_chart()

    def parse_anglerfish_json(self, f):
        """Parse the JSON output from Anglerfish and save the summary statistics"""
        try:
            parsed_json = json.load(f["f"])
        except Exception:
            file = f["fn"]
            log.warning(f"Could not parse Anglerfish JSON: '{file}'")
            return

        # Fetch a sample name from the command
        s_name = f["s_name"]
        self.add_data_source(f, s_name)
        self.anglerfish_data[s_name] = {}

        # Add version info
        self.add_software_version(parsed_json["anglerfish_version"], s_name)

        # Parse Sample Stats
        ## Index for each sample and their reads in order to iterate without knowing sample names
        index = 0
        total_reads = 0
        try:
            for k in parsed_json["sample_stats"]:
                total_reads += float(k.get("#reads", 0))
                for key in k:
                    try:
                        self.anglerfish_data[s_name][f"{key}_{index}"] = float(k[key])
                    except ValueError:
                        self.anglerfish_data[s_name][f"{key}_{index}"] = k[key]
                index += 1
            self.anglerfish_data[s_name]["sample_stats_amount"] = index
            self.anglerfish_data[s_name]["total_read"] = total_reads
        except KeyError:
            # No sample stat in file or sample stat missing info
            self.anglerfish_data[s_name]["sample_stats_amount"] = -1

        # Parse Undetermined Indexes
        ## Index for each undetermined (count, index) pair
        index = 0
        total_count = 0
        try:
            for k in parsed_json["undetermined"]:
                if len(k) > 0:
                    self.anglerfish_data[s_name][f"undetermined_count_{index}"] = float(k["count"])
                    total_count += float(k["count"])
                    self.anglerfish_data[s_name][f"undetermined_index_{index}"] = k["undetermined_index"]
                    index += 1
            self.anglerfish_data[s_name]["undetermined_amount"] = index
        except KeyError:
            # No undetermined in file or undetermined missing info
            self.anglerfish_data[s_name]["undetermined_amount"] = -1
        self.anglerfish_data[s_name]["total_count"] = total_count

    # General stats table
    def anglerfish_general_stats_table(self):
        """Add Anglerfish statistics to the general statistics table"""
        # Prepp data for general stat table
        ## Multiple sample names per file requires dict where the first key is not file name
        data = {}
        for s_name in self.anglerfish_data:
            total_read = self.anglerfish_data[s_name]["total_read"]
            total_count = self.anglerfish_data[s_name]["total_count"]
            try:
                for k in range(self.anglerfish_data[s_name]["sample_stats_amount"]):
                    key = self.anglerfish_data[s_name][f"sample_name_{k}"]
                    data[key] = {}
                    data["undetermined"] = {}
                    data[f"total_read_{s_name}"] = {}
                    reads = self.anglerfish_data[s_name][f"#reads_{k}"]
                    data[key]["#reads"] = reads
                    # data[f"total_read_{s_name}"]["#reads"] = total_read
                    data[key]["mean_read_len"] = self.anglerfish_data[s_name][f"mean_read_len_{k}"]
                    data[key]["std_read_len"] = self.anglerfish_data[s_name][f"std_read_len_{k}"]
                    data[key]["library"] = float((reads / total_read) * 100)
                    data["undetermined"]["library"] = float((total_count / total_read) * 100)
            except KeyError:
                log.debug(f"No general stats table generated from Anglerfish json: {s_name}")
            except ZeroDivisionError:
                log.debug(f"No library in general stats table generated from Anglerfish json: {s_name}")

        headers = {
            "library": {
                "title": "% Library",
                "description": "Fraction within library.",
                "max": 100,
                "min": 0,
                "scale": "PuBu-rev",
                "suffix": " %",
            },
            "#reads": {
                "title": "# Reads",
                "description": "Total number of reads",
                "min": 0,
                "scale": "PuOr",
                "format": "{:.0f}",
            },
            "mean_read_len": {
                "title": "Read Length",
                "description": "Mean read length",
                "min": 0,
                "scale": "RdYlGn",
                "suffix": " bp",
            },
            "std_read_len": {
                "title": "Length StdDev",
                "description": "Standard deviation of the read lengths",
                "min": 0,
                "scale": "RdPu",
                "suffix": " bp",
            },
        }

        self.general_stats_addcols(data, headers)

    def anglerfish_sample_stats(self):
        """Generate plot for read length from sample stats.
        For >10 samples: generate table plot
        for >= 10 samples: generate beeswarm plot"""
        data = {}
        total_samples = 0
        for s_name in self.anglerfish_data:
            index = self.anglerfish_data[s_name]["sample_stats_amount"]
            if index > 0:
                total_samples += index
                for i in range(index):
                    sample_name = self.anglerfish_data[s_name][f"sample_name_{i}"]
                    data[f"Sample: {sample_name}"] = {}
                    data[f"Sample: {sample_name}"]["Mean"] = self.anglerfish_data[s_name][f"mean_read_len_{i}"]
                    data[f"Sample: {sample_name}"]["Standard Deviation"] = self.anglerfish_data[s_name][
                        f"std_read_len_{i}"
                    ]
            else:
                # For non-existing sample stat and faulty sample stat
                log.debug(f"Missing Sample Stat Data in Anglerfish json: {s_name}")
        if len(data) == 0:
            return

        config = {
            "id": "Sample_Stat_Read_Length",
            "title": "Anglerfish: Read Lengths Summary",
        }
        # Plot table if less than 10 samples exist, beeswarm if more
        if total_samples < 10:
            p = table.plot(data, None, config)
        else:
            p = beeswarm.plot(data, None, config)
        self.add_section(
            name="Read Lengths Summary",
            anchor="anglerfish-sample-statistics",
            description="The Mean read length and the Standard Deviation for each sample.",
            plot=p,
        )

    def anglerfish_undetermined_index_chart(self):
        """Generate Undetermined indexes Bar Plot"""
        data = {}
        for s_name in self.anglerfish_data:
            index = self.anglerfish_data[s_name]["undetermined_amount"]
            # Index smaller than 0 caused by KeyError from no undetermined data
            if index > 0:
                for i in range(index):
                    undetermined_index = self.anglerfish_data[s_name][f"undetermined_index_{i}"]
                    data[undetermined_index] = {}
                    data[undetermined_index][undetermined_index] = self.anglerfish_data[s_name][
                        f"undetermined_count_{i}"
                    ]
            else:
                # For non existing undetermined and faulty undetermined
                log.debug(f"Missing Undetermined Data in Anglerfish json: {s_name}")
        # Only add undetermined section if undetermined data exists
        if len(data) == 0:
            return

        config = {
            "id": "Anglerfish_undetermined_index_plot",
            "cpswitch": False,
            "title": "Anglerfish: Undetermined Indexes",
            "ylab": "Index Count",
            "tt_percentages": False,
        }
        self.add_section(
            name="Undetermined Indexes",
            anchor="anglerfish-undetermined-indexes",
            plot=bargraph.plot(data, None, config),
        )



import re
from collections import defaultdict

_R1 = "Read1"
_R2 = "Read2"
_VALID_MATES = [_R1, _R2]


def parse_fastqc_metrics_file(f):
    """
    NA12878.fastqc_metrics.csv

    READ MEAN QUALITY,Read1,Q30 Reads,151104
    ...
    POSITIONAL BASE CONTENT,Read1,ReadPos 1 A Bases,2963930
    ...
    NUCLEOTIDE QUALITY,Read1,Q7 A Bases,777933
    ...
    READ LENGTHS,Read1,145-152bp Length Reads,10142922
    ...
    READ BASE CONTENT,Read1,0% A Reads,1997
    ...
    """
    s_name = re.search(r"(.*).fastqc_metrics.csv", f["fn"]).group(1)
    data_by_sample = initialize_dataset(s_name)

    for line in f["f"].splitlines():
        # This conforms with the standard DRAGEN metrics format
        # Percentage is currently unused
        tokens = line.split(",")
        if len(tokens) == 4:
            group, mate, metric, value = line.split(",")
            percentage = None
        elif len(tokens) == 5:
            group, mate, metric, value, percentage = line.split(",")
        else:
            raise ValueError(f"Unexpected number of values in line {line}")

        try:
            value = int(value)
        except ValueError:
            pass

        # Store each value by group and by metric
        assert mate in _VALID_MATES
        data_by_sample[s_name][mate][group][metric] = value

    # Delete empty mate groups so we don't generate empty datasets
    # Altered to work on python3 which doesn't allow deleting from dict during iterating
    delete = [mate for mate, mate_data in data_by_sample[s_name].items() if len(mate_data) == 0]
    for key in delete:
        del data_by_sample[s_name][key]

    return s_name, data_by_sample


def initialize_dataset(s_name):
    data = dict()
    data[s_name] = dict()
    for mate in _VALID_MATES:
        data[s_name][mate] = defaultdict(lambda: defaultdict(int))
    return data


def average_from_range(metric_range):
    if "+" in metric_range:
        metric_range = metric_range.replace("+", "")
    if ">=" in metric_range:
        metric_range = metric_range.replace(">=", "")
    if "-" in metric_range:
        start, end = metric_range.split("-")
        avg_pos = (int(end) + int(start)) / 2.0
    else:
        avg_pos = int(metric_range)
    return avg_pos


def average_pos_from_metric(metric):
    parts = metric.split()
    metric_range = parts[1]
    return average_from_range(metric_range)


def average_pos_from_size(metric):
    parts = metric.split()
    len_range = parts[0].split("bp")[0]
    return average_from_range(len_range)


def percentage_from_content_metric(metric):
    parts = metric.split()
    pct = int(parts[0].split("%")[0])
    return pct


def pos_qual_table_cmp(key):
    parts = key.split()
    pos = average_from_range(parts[1])
    pct = int(parts[2][:-1])
    return pos * 1000 + pct


def sortPosQualTableKeys(data_dict):
    return sorted(data_dict.keys(), key=pos_qual_table_cmp)



""" MultiQC module to parse logs from Skewer """


import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)

VERSION_REGEX = r"skewer v([\d\.]+) \[.+\]"


class MultiqcModule(BaseMultiqcModule):
    """Skewer"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Skewer",
            anchor="skewer",
            href="https://github.com/relipmoc/skewer",
            info="is an adapter trimming tool specially designed for processing next-generation sequencing (NGS) paired-end sequences.",
            doi="10.1186/1471-2105-15-182",
        )

        self.skewer_data = dict()
        self.skewer_readlen_dist = dict()

        for f in self.find_log_files("skewer", filehandles=True):
            self.parse_skewer_log(f)

        # Filter to strip out ignored sample names
        self.skewer_data = self.ignore_samples(self.skewer_data)

        if len(self.skewer_data) == 0:
            raise ModuleNoSamplesFound

        headers = {
            "pct_trimmed": {
                "title": "% Trimmed",
                "description": "% of reads trimmed",
                "scale": "RdYlGn-rev",
                "max": 100,
                "min": 0,
                "suffix": "%",
            }
        }

        self.general_stats_addcols(self.skewer_data, headers)

        # Write parsed report data to a file
        self.write_data_file(self.skewer_data, "multiqc_skewer")
        self.write_data_file(self.skewer_readlen_dist, "multiqc_skewer_readlen_dist")

        # set the value 0 for every x where a given sample doens't have a value
        all_x_values = []
        for s_name in self.skewer_readlen_dist:
            for xval in self.skewer_readlen_dist[s_name]:
                all_x_values.append(xval)

        for s_name in self.skewer_readlen_dist:
            for xval in all_x_values:
                if xval not in self.skewer_readlen_dist[s_name]:
                    self.skewer_readlen_dist[s_name][xval] = 0.0

        # add the histogram to the report
        self.add_readlen_dist_plot()

        log.info(f"Found {len(self.skewer_data)} reports")

    def add_readlen_dist_plot(self):
        """Generate plot HTML for read length distribution plot."""
        pconfig = {
            "id": "skewer_read_length_histogram",
            "title": "Skewer: Read Length Distribution after trimming",
            "xDecimals": False,
            "ylab": "% of Reads",
            "xlab": "Read Length",
            "xmin": 0,
            "ymin": 0,
            "tt_label": "<b>{point.x}</b>: {point.y:.1f}%",
        }
        self.add_section(plot=linegraph.plot(self.skewer_readlen_dist, pconfig))

    def parse_skewer_log(self, f):
        """Go through log file looking for skewer output"""
        fh = f["f"]
        regexes = {
            "fq1": r"Input file:\s+(.+)",
            "fq2": r"Paired file:\s+(.+)",
            "r_processed": r"(\d+) read|reads pairs? processed",
            "r_short_filtered": r"(\d+) \(\s*\d+.\d+%\) short read",
            "r_empty_filtered": r"(\d+) \(\s*\d+.\d+%\) empty read",
            "r_avail": r"(\d+) \(\s*\d+.\d+%\) read",
            "r_trimmed": r"(\d+) \(\s*\d+.\d+%\) trimmed read",
            "r_untrimmed": r"(\d+) \(\s*\d+.\d+%\) untrimmed read",
        }
        regex_hist = r"\s?(\d+)\s+(\d+)\s+(\d+.\d+)%"

        data = dict()
        for k, v in regexes.items():
            data[k] = 0
        data["fq1"] = None
        data["fq2"] = None
        readlen_dist = dict()

        for line in fh:
            if line.startswith("skewer"):
                match = re.search(VERSION_REGEX, line)
                if match:
                    data["version"] = match.group(1)

            for k, r in regexes.items():
                match = re.search(r, line)
                if match:
                    data[k] = match.group(1).replace(",", "")

            match = re.search(regex_hist, line)
            if match:
                read_length = int(match.group(1))
                pct_at_rl = float(match.group(3))
                readlen_dist[read_length] = pct_at_rl

        if data["fq1"] is not None:
            s_name = self.clean_s_name(data["fq1"], f)
            if s_name in self.skewer_readlen_dist:
                log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
            self.add_data_source(f, s_name)
            self.add_skewer_data(s_name, data, f)
            self.skewer_readlen_dist[s_name] = readlen_dist
            self.add_software_version(data.get("version"), s_name)

        if data["fq2"] is not None:
            s_name = self.clean_s_name(data["fq1"], f)
            if s_name in self.skewer_readlen_dist:
                log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
            self.add_data_source(f, s_name)
            self.add_skewer_data(s_name, data, f)
            self.skewer_readlen_dist[s_name] = readlen_dist
            self.add_software_version(data.get("version"), s_name)

    def add_skewer_data(self, s_name, data, f):
        stats = ["r_processed", "r_short_filtered", "r_empty_filtered", "r_avail", "r_trimmed", "r_untrimmed"]
        if s_name in self.skewer_data:
            log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
        self.skewer_data[s_name] = {}
        self.add_data_source(f, s_name)
        for k in stats:
            self.skewer_data[s_name][k] = int(data[k])

        self.skewer_data[s_name]["pct_avail"] = (
            100.0 * float(data["r_avail"]) / float(data["r_processed"]) if float(data["r_processed"]) else None
        )
        self.skewer_data[s_name]["pct_trimmed"] = (
            100.0 * float(data["r_trimmed"]) / float(data["r_avail"]) if float(data["r_avail"]) else None
        )
        self.skewer_data[s_name]["pct_untrimmed"] = (
            100.0 * float(data["r_untrimmed"]) / float(data["r_avail"]) if float(data["r_avail"]) else None
        )



#!/usr/bin/env python

""" Utility functions to handle software version reporting """


import logging
import os
from collections import defaultdict
from typing import List, Dict

import packaging.version
import yaml

from multiqc.utils import report as mqc_report

# Initialise the logger
log = logging.getLogger(__name__)


def normalize_name(name: str):
    """Normalize software name to lowercase and remove spaces, underscores and dashes"""
    return name.lower().replace(" ", "").replace("-", "").replace("_", "")


def update_versions_from_config(config, report):
    """Update report with software versions from config if provided"""
    # Parse software version from config if provided
    versions_from_config = load_versions_from_config(config)
    for group, softwares in versions_from_config.items():
        # Try to find if the software is listed among the executed modules.
        # Unlisted software are still reported in the `Software Versions` section.
        module = find_matching_module(group, report.modules_output)

        # Map normalized module software names to the nicely formatted names.
        module_softwares = {}
        if module is not None:
            module_softwares = {normalize_name(m_software): m_software for m_software in module.versions}

            # Use the nicely formatted module name as group name
            group = module.name

        for software, versions in softwares.items():
            # Update versions if the software is listed among the executed modules
            if module is not None and not config.disable_version_detection:
                # Update software name to match the module name format if found
                software = module_softwares.get(normalize_name(software), software)

                for version in versions:
                    module.add_software_version(str(version), software_name=software)

                # Get the updated software versions from the module
                versions = module.versions[software]

            # Add updated software versions to the report
            report.software_versions[group][software] = versions


def load_versions_from_config(config):
    """Try to load software versions from config"""
    log.debug("Reading software versions from config.software_versions")
    versions_config = getattr(config, "software_versions", defaultdict(lambda: defaultdict(list)))
    if not isinstance(versions_config, dict):
        log.error("Expected the `software_versions` config section to be a dictionary")
        versions_config = {}
    else:
        versions_config = validate_software_versions(versions_config)

    versions_from_files = defaultdict(lambda: defaultdict(list))
    for f in mqc_report.files.get("software_versions", []):
        file_name = os.path.join(f["root"], f["fn"])
        with open(file_name) as fh:
            log.debug(f"Reading software versions settings from: {file_name}")
            try:
                versions_from_one_file = yaml.load(
                    fh,
                    # We need to be cautious when loading unquoted version strings from a YAML file.
                    # For instance, the version `1.10` will be parsed as a float by default, this converted
                    # into `1.1`. Passing yaml.BaseLoader explicitly makes YAML treat all scalar values
                    # as strings, so `1.10` will turn into a string `"1.10"` as we want.
                    # From https://pyyaml.org/wiki/PyYAMLDocumentation
                    #      BaseLoader(stream) does not resolve or support any tags
                    #      and constructs only basic Python objects: lists, dictionaries and Unicode strings.
                    Loader=yaml.BaseLoader,
                )
            except yaml.scanner.ScannerError as e:
                log.error(f"Error parsing versions YAML: {e}")

        if not isinstance(versions_from_one_file, dict):
            log.error(
                f"Expected the software versions file {file_name} to contain a dictionary structure, "
                f"ignoring the file."
            )
            continue
        versions_from_one_file = validate_software_versions(versions_from_one_file)
        config.update_dict(versions_from_files, versions_from_one_file)

    # Aggregate versions listed in config and file
    config.update_dict(versions_config, versions_from_files)

    # Parse the aggregated versions
    for group in versions_config:
        softwares = versions_config[group]
        for tool in softwares:
            # Try and convert version to packaging.versions.Version object and remove duplicates
            versions = list(set([parse_version(version) for version in softwares[tool]]))
            versions = sort_versions(versions)
            softwares[tool] = versions

    return versions_config


def validate_software_versions(versions_config: Dict) -> Dict[str, Dict]:
    """
    Validate software versions input from config file

    Two main formats are supported:

    (1) A dict of dicts, where the first level is the group name and the second level is the software name.
        The software name maps to a list of versions or a single version string.

        group_name:
            software_name:
                - version1
                - version2
        other_group:
            other_software: version3

    (2) A dict of lists, where the key is the software name and the value is a list of versions or a single version string.
        In this case the group name is set to the software name.

        software_name:
            - version1
            - version2
        other_software: version3

    It is also possible to mix the two formats, but this is not recommended.

    Returns a dict of dicts of list in the format (1).
    """

    def _filter_list(lst: List[str]) -> List[str]:
        """Remove all non-string version tags"""
        fixed_lst = []
        for item in lst:
            if not isinstance(item, str):
                log.error(
                    f"Version must be a string, got '{type(item).__name__}': '{item}'. Consider wrapping the value in quotes: '\"{item}\"'"
                )
            else:
                fixed_lst.append(item)
        return fixed_lst

    output = defaultdict(lambda: defaultdict(list))

    for level1_key, level1_values in versions_config.items():
        group = level1_key
        software = level1_key

        # Check if the input is in format (1)
        if isinstance(level1_values, dict):
            for level2_key, versions in level1_values.items():
                software = level2_key
                if not isinstance(versions, list):
                    versions = [versions]
                versions = _filter_list(versions)
                if versions:
                    output[group][software] = versions

        # Check if the input is in format (2)
        else:
            versions = level1_values
            if not isinstance(versions, list):
                versions = [versions]
            versions = _filter_list(versions)
            if versions:
                output[group][software] = versions

    return output


def sort_versions(versions):
    """
    Sort list of versions in descending order. Accepts list with both strings and packaging.version.Version
    objects.
    """
    version_objs = [v for v in versions if isinstance(v, packaging.version.Version)]
    version_strs = [v for v in versions if not isinstance(v, packaging.version.Version)]
    versions = sorted(version_objs) + sorted(version_strs)
    return versions


def find_matching_module(software_name: str, modules):
    """
    Find the module by name
    """
    d = {normalize_name(m.name): m for m in modules}
    return d.get(normalize_name(software_name))


def parse_version(version: str):
    """
    Check if version string is PEP 440 compliant to enable version normalization and proper ordering.
    Returns tuple with version and a boolean indicating if version is PEP 440 compliant.
    # - https://peps.python.org/pep-0440/
    """
    try:
        return packaging.version.parse(version)
    except packaging.version.InvalidVersion:
        return str(version)
